//**************************************************************************//
//**                                                                      **//
//**                            WinNES                                    **//
//**       Nintendo Entertainment System Emulator for Windows             **//
//**                                                                      **//
//**   Copyright (C) Dott. Cosimo De Michele 1998 / Sremulador 2005       **//
//**   E-mail : karakiri@hotmail.com, Sremulador@bol.com.br               **//
//**         You are not allowed to distribuite this software             **//
//**         commercially. Please, notify me, if you make any             **//
//**         changes to this file.                                        **//
//**                                                                      **//
//**   Nintendo is a trademark of Nintendo inc.                            **//
//**************************************************************************//

unit Nesmapper;

interface

const
  IRQon = true;

type
  TMMCWrite = procedure (Addr : word; value : byte);
  { MAPPER ROUTINE }
  procedure VROM_BANK1(A : word; V : byte);
  procedure VROM_BANK4(A : word; V : byte);
  procedure VROM_BANK8(A : word; V : byte);
  procedure ROM_BANK8 (A : word; V : byte);
  procedure ROM_BANK16(A : word; V : byte);
  procedure ROM_BANK32(A : word; V : byte);
  { NO MAPPER }
  function  MMCInit(MMCType : integer): boolean;
  procedure NONEInit;
  procedure NONEWrite(Addr : word; value : byte);
  procedure IRQz(var Result: byte);

var
  Mappers  : string;
  MMCWrite : TMMCWrite;
  IRQCount, IRQLatch: word;
  IRQa: boolean;
  { MMC1 }
  MMC1_Reg : array[0..3] of byte;
  MMC1_sft,
  MMC1_buf : byte;
  MMC1_MIRRORING : word;
  BIGBANK: integer;
  MMC2_latch,
  MMC3_cmd,
  MMC3_Chr,
  MMC4_latch1,
  MMC4_latch2 : byte;
  cbase: word;
  IRQ_Counter : word;
  IRQ_ratch : byte;
  IRQ_flg : boolean;
  vmask : integer;

implementation

uses neS, SysUtils, dialogs;

  procedure MIRROR_SET(V : byte);
  begin
    NES.MirrorXor := $800 shr V;
    NES.Mirroring := V;
  end; { MIRROR_SET }

  procedure VROM_BANK1(A : word; V : byte);
  begin
      if VRomSize>0 then
       VPage[(A shr 10)] := pMemory(pChar(NES.VROM) + (V * $400) - A)
      else
       VPage[(A shr 10)] := pMemory(pChar(@VRAM) + (V * $400) - A);
  end; { VROM_BANK1 }

  procedure VROM_BANK4(A : word; V : byte);
  begin
      if VRomSize>0 then
      begin
        VPage[(A shr 10) + 0] := pMemory(pChar(VROM) + (V * $1000) - A);
        VPage[(A shr 10) + 1] := pMemory(pChar(VROM) + (V * $1000) - A);
        VPage[(A shr 10) + 2] := pMemory(pChar(VROM) + (V * $1000) - A);
        VPage[(A shr 10) + 3] := pMemory(pChar(VROM) + (V * $1000) - A);
      end
      else
      begin
        VPage[(A shr 10) + 0] := pMemory(pChar(@VRAM) + (V * $1000) - A);
        VPage[(A shr 10) + 1] := pMemory(pChar(@VRAM) + (V * $1000) - A);
        VPage[(A shr 10) + 2] := pMemory(pChar(@VRAM) + (V * $1000) - A);
        VPage[(A shr 10) + 3] := pMemory(pChar(@VRAM) + (V * $1000) - A);
      end;
  end; { VROM_BANK4 }

  procedure VROM_BANK8(A : word; V : byte);
  begin
      if VRomSize>0 then
      begin
        VPage[0] := pMemory(pChar(VROM) + (V * $2000) - A);
        VPage[1] := pMemory(pChar(VROM) + (V * $2000) - A);
        VPage[2] := pMemory(pChar(VROM) + (V * $2000) - A);
        VPage[3] := pMemory(pChar(VROM) + (V * $2000) - A);
        VPage[4] := pMemory(pChar(VROM) + (V * $2000) - A);
        VPage[5] := pMemory(pChar(VROM) + (V * $2000) - A);
        VPage[6] := pMemory(pChar(VROM) + (V * $2000) - A);
        VPage[7] := pMemory(pChar(VROM) + (V * $2000) - A);
      end
      else
      begin
        VPage[0] := pMemory(pChar(@VRAM) + (V * $2000) - A);
        VPage[1] := pMemory(pChar(@VRAM) + (V * $2000) - A);
        VPage[2] := pMemory(pChar(@VRAM) + (V * $2000) - A);
        VPage[3] := pMemory(pChar(@VRAM) + (V * $2000) - A);
        VPage[4] := pMemory(pChar(@VRAM) + (V * $2000) - A);
        VPage[5] := pMemory(pChar(@VRAM) + (V * $2000) - A);
        VPage[6] := pMemory(pChar(@VRAM) + (V * $2000) - A);
        VPage[7] := pMemory(pChar(@VRAM) + (V * $2000) - A);
      end;
  end; { VROM_BANK8 }

  procedure ROM_BANK8(A : word; V : byte);
  begin
    NES.Page[(A shr 13)] := pMemory(pChar(NES.ROM) + (V * $2000) - A);
  end; { ROM_BANK8 }

  procedure ROM_BANK16(A : word; V : byte);
  begin
      Page[(A shr 13) + 0] := pMemory(pChar(ROM) + (V * $4000) - A);
      Page[(A shr 13) + 1] := pMemory(pChar(ROM) + (V * $4000) - A);
  end; { ROM_BANK16 }

  procedure ROM_BANK32(A : word; V : byte);
  begin
      Page[4] := pMemory(pChar(ROM) + (V * $8000) - A);
      Page[5] := pMemory(pChar(ROM) + (V * $8000) - A);
      Page[6] := pMemory(pChar(ROM) + (V * $8000) - A);
      Page[7] := pMemory(pChar(ROM) + (V * $8000) - A);
  end; { ROM_BANK32 }

  { ***************************** }
  { * NONE MAPPER               * }
  { ***************************** }
  procedure NONEInit;
  var
    i : integer;
  begin
      ROM_BANK16($8000, 0);
      ROM_BANK16($C000, (RomSize - 1));
      for i:= Low(VPage) to High(VPage) do
        if VROMSize>0 then
          VPage[i] := VROM
        else
          VPage[i] := @VRAM;
      vmask := (VRomSize - 1);
  end; { NONEInit }

  procedure NONEWrite(Addr : word; value : byte);
  begin
  end; { NONEWrite }

  procedure IRQz(var Result: byte);
  begin
    case MMCType of
     118,
      99,
      65,
      16,
      17,
      22,
      4   : begin
              if (UCount)=0 then
              begin
                if ((IRQa) and (IRQon)) then
                begin
                  dec(IRQCount);
                  if (IRQCount<=0) then
                  begin
                    if (MMCType=65) then
                      IRQCount := IRQLatch;
                    IRQa := false;
                    PPUSCROLL_Change := true;
//                    Result := INT_IRQ;
                  end;
                end;
              end;
            end;
      24,
      85  : begin
              if (UCount=0) then
              begin
		if ((IRQa) and (IRQon)) then
		begin
		  inc(IRQCount);
		  if (IRQCount>=$100) then
                  begin
		    IRQa := false;
		    IRQCount := 0;
//		    Result := INT_IRQ;
                  end;
                end;
              end;
            end;
    end;
  end; { IRQz }

  { ***************************** }
  { * 1 : NINTENDO MMC1         * }
  { ***************************** }
  procedure MMC1Init;
  var
    LastBank, i : integer;
  begin
//    LastBank:= (frmNES.RomSize - 1);
    for i:=0 to 3 do
        MMC1_Reg[i] := 0;
    MMC1_sft := 0;
    MMC1_buf := 0;
    MMC1_MIRRORING := $000;
    if(((NES.ROMType and ROM_MIRROR)<>0) and (NES.VROMSize=0)) then
      MMC1_MIRRORING := $800;
  end; { MMC1Init }

  procedure MMC1Write(Addr : word; value : byte);
  var
    n : integer;
  begin
    n := (Addr shr 13) - 4;
    BIGBANK := 0;
    if (value and $80)<>0 then  begin
      MMC1_sft := 0;
      MMC1_buf := 0;
      MMC1_Reg[0] := MMC1_Reg[0] or $0C;
      Exit;
    end;
    MMC1_buf := MMC1_buf or ((value and 1) shl MMC1_sft);
    inc(MMC1_sft);
    if MMC1_sft=5 then begin
      MMC1_Reg[n] := MMC1_buf;
      value := MMC1_buf;
      MMC1_sft := 0;
      MMC1_buf := 0;
      case n of
        0 : begin
                MIRROR_SET(value and ROM_MIRROR);
                if (value and 2)<>0 then
                  MirrorXor := MirrorXor + MMC1_MIRRORING * (value and ROM_MIRROR);
            end;
        1 : begin
                if (VROMSize=0) then
                begin
                  if (value and $10)<>0 then
                    BIGBANK := 1
                  else
                    BIGBANK := 0;
                end
                else
                begin
                  value := value and $1F;
                  if (MMC1_Reg[0] and $10)<>0 then
                    VROM_BANK4($0000, value)
                  else
                    VROM_BANK8($0000, (Value shr 1));
                end;
            end;
        2 : begin
                if (MMC1_Reg[0] and $10)<>0 then
                  VROM_BANK4($1000, Value);
            end;
        3 : begin
                value := value and $F;
                if (MMC1_Reg[0] and 8)<>0 then
                begin
                  if ((MMC1_Reg[0] and 4)<>0) then
                  begin
                    if BIGBANK<>0 then
                      ROM_BANK16($8000, Value + $10*BIGBANK)
                    else
                      ROM_BANK16($8000, Value)
                  end
                  else
                  begin
                    if BIGBANK<>0 then
                      ROM_BANK16($C000, Value + $10*BIGBANK)
                    else
                      ROM_BANK16($C000, Value);
                  end
                end
                else
                  ROM_BANK32((value shr 1) * $8000, (Value shr 1));
            end;
    end;
    end;
  end; { MMC1Write }

  { ***************************** }
  { * 2 : 74HC161/74HC32        * }
  { ***************************** }
  procedure iNES2Write(Addr : word; value : byte);
  begin
    //if (Addr>$8000) and (Addr<$FFFF) then
      ROM_BANK16($8000, value);
  end; { iNES2Write }

  { ***************************** }
  { * 3 : VROM switch           * }
  { ***************************** }
  procedure VROMswWrite(Addr : word; value : byte);
  begin
    //if (Addr>$8000) and (Addr<$FFFF) then
      VROM_BANK8($0000, value and vmask);
  end; { VROMswWrite }

  { ***************************** }
  { * 4 : NINTENDO MMC3         * }
  { ***************************** }
  procedure MMC3Init;
  begin
    cbase := 0;
  end; { MMC3Init }

  procedure MMC3Write(Addr : word; value : byte);
  begin
    case (Addr and $E001) of
      $8000 : begin
                MMC3_cmd := value;
                if (MMC3_cmd and $80)<>0 then
                  cbase := $1000
                else
                  cbase := 0;
              end;
      $8001 : case (MMC3_cmd and $7) of
                0 : begin
                      VROM_BANK1(cbase xor $0000, value);
                      VROM_BANK1(cbase xor $0400, value + 1);
                    end;
                1 : begin
                      VROM_BANK1(cbase xor $0800, value);
                      VROM_BANK1(cbase xor $0C00, value + 1);
                    end;
                2 : VROM_BANK1(cbase xor $1000, value);
                3 : VROM_BANK1(cbase xor $1400, value);
                4 : VROM_BANK1(cbase xor $1800, value);
                5 : VROM_BANK1(cbase xor $1C00, value);
                6 : if (MMC3_cmd and $40)<>0 then
                    begin
                      ROM_BANK8($8000, (NES.RomSize - 1) * 2);
                      ROM_BANK8($C000, value);
                    end
                    else
                    begin
                      ROM_BANK8($C000, (NES.RomSize - 1) * 2);
                      ROM_BANK8($8000, value);
                    end;
                7 : if (MMC3_cmd and $40)<>0 then
                    begin
                      ROM_BANK8($8000, (NES.RomSize - 1) * 2);
                      ROM_BANK8($A000, value);
                    end
                    else
                    begin
                      ROM_BANK8($C000, (NES.RomSize - 1) * 2);
                      ROM_BANK8($A000, value);
                    end;
              end;
      $A000 : MIRROR_SET(value and ROM_MIRROR);
      $A001 : if (value and $80)<>0 then
                NES.ROMType := NES.ROMType or ROM_WRAM
              else NES.ROMType := NES.ROMType and not(ROM_WRAM);
      $C000 : begin
                if((Value=0) and (IRQa))then
                  IRQCount := IRQLatch
                else
                  IRQCount := value;
              end;
      $C001 : begin
                IRQLatch := Value;
                IRQa := true;
              end;
      $C002 : IRQa := false;
      $C003 : IRQa := false;
{      $E000 : begin
              IRQa := false;
//              IRQCount := IRQLatch;
              end;
      // Corrigge bug no sceen false}
      $E000,
      $E003 : IRQa := false;
      $E001,
      $E002 : IRQa := False;
    end;
  end; { MMC3Write }


  { ***************************** }
  { * 5 : FFE F3xxx             * }
  { ***************************** }


 procedure MMC5Init;
  begin
    cbase := 0;
  end; { MMC5Init }


  //trabalhando no mapa

  procedure MMC5Write(Addr : word; value : byte);
  begin
    case (Addr and $E001) of
      $5010 : begin
              IRQa := true;

              end;
      $5115 : VROM_BANK1(cbase xor $8000, value);
      $5116 : VROM_BANK1(cbase xor $C000, value);
    end;
  end; { MMC3Write }

  { ***************************** }
  { * 7 : ROM Switch            * }
  { ***************************** }
  procedure ROMswInit;
  begin
    ROM_BANK32($8000, 0);
  end; { ROMswInit }

  procedure ROMswWrite(Addr : word; value : byte);
  begin
    NES.Mirrorxor := $400 shl (Value shr 4);
    ROM_BANK32($8000,  value and $F);
  end; { ROMswWrite }

  { ***************************** }
  { * 8 : FFE F3xxx             * }
  { ***************************** }
  procedure FFEF3Init;
  begin
    ROM_BANK16($8000, 0);
    ROM_BANK16($C000, 1);
  end; { FFEF3Init }

  procedure FFEF3Write(Addr : word; value : byte);
  begin
    ROM_BANK16($8000, value shr $3);
    VROM_BANK8($0000, value and $7)
  end; { FFEF3Write }

  { ***************************** }
  { * 9 : NINTENDO MMC2         * }
  { ***************************** }
  procedure MMC2Init;
  begin
    MMC2_latch := $FE;
    ROM_BANK8($A000,(NES.RomSize * 2) - 2);
    ROM_BANK8($C000,(NES.RomSize * 2));
    ROM_BANK8($E000,(NES.RomSize * 2) + 1);
//    ROM_BANK16($8000, frmNES.RomSize - 2);
//    ROM_BANK16($C000, frmNES.RomSize - 1);
    ROM_BANK8($8000, 0);
  end; { MMC2Init }

  procedure MMC2Write(Addr : word; value : byte);
  begin
    case (Addr and $F000) of
      $A000 : ROM_BANK8($8000, value);
      $B000,
      $C000 : VROM_BANK4($0000, value);
      $D000 : if (MMC2_latch=$FD) then
                  VROM_BANK4($1000, value);
      $E000 : if (MMC2_latch=$FE) then
                  VROM_BANK4($1000, value);
      $F000 : MIRROR_SET(value and ROM_MIRROR);
    end;
  end; { MMC2Write }

  { ***************************** }
  { * 10 : NINTENDO MMC4        * }
  { ***************************** }
  procedure MMC4Init;
  begin
    MMC4_latch1 := $FE;
    MMC4_latch2 := $FE;
  end; { MMC4Init }

  procedure MMC4Write(Addr : word; value : byte);
  begin
    case (Addr and $F000) of
      $A000 : ROM_BANK16($8000, value);
      $B000 : if (MMC4_latch1=$FD) then
                VROM_BANK4($0000, value);
      $C000 : if (MMC4_latch1=$FE) then
                VROM_BANK4($0000, value);
      $D000 : if (MMC4_latch2=$FD) then
                VROM_BANK4($1000, value);
      $E000 : if (MMC4_latch2=$FE) then
                VROM_BANK4($1000, value);
      $F000 : MIRROR_SET(value and ROM_MIRROR);
    end;
  end; { MMC4Write }

  { ***************************** }
  { * 11 : Color dream          * }
  { ***************************** }
  procedure iNES11Init;
  begin
    ROM_BANK32($8000, 0);
    VROM_BANK8($0000, 0);
  end; { iNES11Init }

  procedure iNES11Write(Addr : word; value : byte);
  begin
    
    ROM_BANK32($8000, value and $15);
  	VROM_BANK8($0000, Value shr 4);
  end; { iNES11Write }

  { ***************************** }
  { * 15 : 100-in-1             * }
  { ***************************** }
  procedure iNES100in1Init;
  begin
    ROM_BANK32($8000, 0);
  end; { iNES100in1Init }

  procedure iNES100in1Write(Addr : word; value : byte);
  begin
    case (Addr) of
      $8000 : begin
                MIRROR_SET(((value shr 6) and ROM_MIRROR) xor ROM_MIRROR);
                value := value and $3F;
                ROM_BANK16($8000, Value);
		ROM_BANK16($C000, Value + 1);
              end;
      $8001 : begin
                value := value and $3F;
                ROM_BANK16($C000, value);
              end;
      $8002 : begin
                if (value and $80)<>0 then
                begin
                  ROM_BANK8($8000, (Value * 2) + 1);
                  ROM_BANK8($A000, (Value * 2) + 1);
                  ROM_BANK8($C000, (Value * 2) + 1);
                  ROM_BANK8($E000, (Value * 2) + 1);
                end
                else
                begin
                  ROM_BANK8($8000, (Value * 2));
                  ROM_BANK8($A000, (Value * 2));
                  ROM_BANK8($C000, (Value * 2));
                  ROM_BANK8($E000, (Value * 2));
                end;
              end;
      $8003 : begin
                MIRROR_SET(((value shr 6) and ROM_MIRROR) xor ROM_MIRROR);
                value := value and $3F;
                ROM_BANK16($C000, value);
              end;
      $F001: begin
      		if (Value and $80)<>0 then
                begin
		  value := value and $3F;
		  ROM_BANK8($C000, value * 2 + 1);
		  ROM_BANK8($E000, value * 2);
                end
		else
		  ROM_BANK16($C000, value);
              end;
    end;
  end; { iNES100in1Write }

  { ***************************** }
  { * 16 : Bandai               * }
  { ***************************** }
  procedure BandaiInit;
  begin
    IRQ_flg := false;
    IRQ_Counter := 0;
  end; { BandaiInit }

  procedure BandaiWrite(Addr : word; value : byte);
  begin
    case (Addr and $F) of
      $0 : VROM_BANK1($0000, value);
      $1 : VROM_BANK1($0400, value);
      $2 : VROM_BANK1($0800, value);
      $3 : VROM_BANK1($0C00, value);
      $4 : VROM_BANK1($1000, value);
      $5 : VROM_BANK1($1400, value);
      $6 : VROM_BANK1($1800, value);
      $7 : VROM_BANK1($1C00, value);
      $8 : ROM_BANK16($8000, value);
      $9 : if (value and ROM_MIRROR)=0 then
             NES.Mirrorxor := $800
           else
	     NES.Mirrorxor := $400;
      $A : IRQ_flg := (Value and 1)<>0;
      //$B : IRQ_counter := (IRQ_counter and $FF00) or value;
      $C : begin
             IRQ_counter := value;
             IRQ_flg := true;
           end
    end;
  end; { BandaiWrite }

  { ***************************** }
  { *    MMC INIT               * }
  { ***************************** }
  function MMCInit(MMCType : integer): boolean;
  begin
    IRQ_flg := false;
    IRQ_counter := 0;
    Result := true;
    case MMCType of
  0, 22 : begin
            Mappers := 'No Mapper';
            NONEInit;
            MMCWrite := NONEWrite;
          end;
      1 : begin
            Mappers := 'Nintendo MMC1';
            NONEInit;
            MMC1Init;
            MMCWrite := MMC1Write;
          end;
      2 : begin
            Mappers := '74HC161/74HC32';
            NONEInit;
            MMCWrite := iNES2Write;
          end;
      3 : begin
            Mappers := 'VROM Switch';
            NONEInit;
            MMCWrite := VROMswWrite;
          end;
  4,118 : begin
            Mappers := 'Nintendo MMC3';
            NONEInit;
            MMC3Init;
            MMCWrite := MMC3Write;
          end;
      5 : begin
            Mappers := 'Nintendo MMC5';
            NONEInit;
            MMC5Init;
            MMCWrite := MMC5Write;
          end;
      7 : begin
            Mappers := 'Rom Switch';
            NONEInit;
            ROMswInit;
            MMCWrite := ROMswWrite;
          end;
      8 : begin
            Mappers := 'FFE F3xxxx';
            NONEInit;
            FFEF3Init;
            MMCWrite := FFEF3Write;
          end;
      9 : begin
            Mappers := 'Nintendo MMC2';
            NONEInit;
            MMC2Init;
            MMCWrite := MMC2Write;
          end;
      10: begin
            Mappers := 'Nintendo MMC4';
            NONEInit;
            MMC4Init;
            MMCWrite := MMC4Write;
          end;
      11: begin
            Mappers := 'Color Dreams Mapper';
            NONEInit;
            iNES11Init;
            MMCWrite := iNES11Write;
          end;
      15: begin
            Mappers := '100-in-1';
            NONEInit;
            iNES100in1Init;
            MMCWrite := iNES100in1Write;
          end;
      16: begin
            Mappers := 'Bandai';
            NONEInit;
            BandaiInit;
            MMCWrite := BandaiWrite;
          end;

    else
//      NES.Terminal.Lines.Add('* iNES mapper #' + IntToStr(MMCType) + ' not supported.');
      Result := false;
    end;
  end; { MMCInit }
(*
	{ 5, "MMC5",                  Mapper5_Write },
	{ 18, "Jaleco SS8806",        Mapper18_Write },
	{ 25, "Konami VRC 2/4 Even",  Mapper25_Write },
	{ 33, "Taito TC0190",         Mapper33_Write },
	{ 34, "Nina-1",               Mapper34_Write },
	{ 64, "Rambo",                Mapper64_Write },
	{ 65, "Irem H3001",           Mapper65_Write },
	{ 66, "74161/32 Jaleco",      Mapper66_Write },
	{ 68, "SunSoft 4",            Mapper68_Write },
	{ 69, "SunSoft 5",            NoneWR },
	{ 71, "Camerica",             Mapper71_Write },
	{ 78, "74161/32 Irem",        Mapper78_Write },
	{ 79, "Nina-3 (AVE)",         Mapper79_Write },
	{ -1, "Not Supported",        NoneWR },
 *)

end.

