unit coleco;

interface
uses sdl,z80,constantes,windows,entradas,tms99xx,arcade_const,sn_76496,timers,zipforge,sysutils,zlibex,dialogs,lenguaje;

procedure Cargar_coleco;
procedure coleco_principal;pascal;
function coleco_getbyte(direccion:word): Byte; pascal;
procedure coleco_putbyte(direccion:word;valor:byte);pascal;
function coleco_inbyte(puerto:word):byte;pascal;
procedure coleco_outbyte(valor:byte;puerto:word);pascal;
function iniciar_coleco:boolean;pascal;
procedure reset_coleco;pascal;
procedure cerrar_coleco;pascal;
procedure coleco_despues_instruccion;pascal;
procedure abrir_coleco;pascal;
procedure coleco_interrupt(int:boolean);pascal;
procedure coleco_grabar_snapshot;pascal;

const
        coleco_bios:tipo_roms=(n:'coleco.rom';l:$2000;p:0);
        keycodes: array[0..15] of byte =
        ($0A,$0D,$07,$0C,$02,$03,$0E,$05,$01,$0B,$06,$09,$0F,$0F,$0F,$0F);
var
  njoymode:boolean;
  nJoyState: array[0..1] of Integer = (0, $FFFF);

implementation
uses principal;

procedure Cargar_coleco;
begin
form1.Panel2.Visible:=true;
form1.BitBtn9.visible:=false;
form1.BitBtn10.Enabled:=true;
form1.BitBtn10.Glyph:=nil;
form1.imagelist2.GetBitmap(4,form1.BitBtn10.Glyph);
form1.BitBtn10.visible:=true;
form1.BitBtn10.OnClick:=form1.fLoadCartucho;
form1.BitBtn11.visible:=true;
form1.BitBtn11.Enabled:=true;
form1.BitBtn12.visible:=false;
form1.BitBtn14.visible:=false;
form1.colecovision1.Checked:=true;
llamadas_maquina.caption:='ColecoVision';
llamadas_maquina.iniciar:=iniciar_coleco;
llamadas_maquina.bucle_general:=coleco_principal;
llamadas_maquina.cerrar:=cerrar_coleco;
llamadas_maquina.reset:=reset_coleco;
llamadas_maquina.cartuchos:=abrir_coleco;
llamadas_maquina.canales_usados:=1;
llamadas_maquina.grabar_snapshot:=coleco_grabar_snapshot;
llamadas_maquina.fps_max:=10738635/2/342/262;
end;

procedure cerrar_coleco;pascal;
begin
SN76496_close(0);
cerrar_cpus;
cerrar_audio;
cerrar_video;
end;

procedure reset_coleco;pascal;
begin
 njoymode:=false;
 SN76496_reset(0);
 TMS99XX_reset;
 reset_z80(main_z80);
 reset_audio;
end;

function iniciar_coleco:boolean;pascal;
begin
iniciar_coleco:=false;
iniciar_audio(1);
//Pantallas:  principal+char y sprites
p_final[0].x:=256+BORDER*2;
p_final[0].y:=192+BORDER*2;
iniciar_video;
//Main CPU
init_cpu(z80_main,0,3579545/llamadas_maquina.fps_max);
main_z80.daisy:=false;
main_z80_ll.getbyte:=coleco_getbyte;
main_z80_ll.putbyte:=coleco_putbyte;
main_z80_ll.outbyte:=coleco_outbyte;
main_z80_ll.inbyte:=coleco_inbyte;
{cargar roms}
if not(cargar_roms(@memoria[0],@coleco_bios,'\coleco.zip',1)) then exit;
//TMS
TMS99XX_Init;
tms.IRQ_Handler:=coleco_interrupt;
//Chip Sonido
SN76496_init(0,3579545);
//Timers
init_timer(0,0,3579545/44100,coleco_despues_instruccion,true);
//final
reset_coleco;
iniciar_coleco:=true;
end;

procedure coleco_principal;pascal;
var
  frame:single;
begin
iniciar_entradas;
frame:=0;
while EmuStatus=EsRuning do begin
  frame:=frame+main_z80.tframes;
  run_z80(main_z80,main_z80_ll,trunc(frame));
  frame:=frame-main_z80.contador;
  main_z80.contador:=0;
  TMS99XX_Interrupt;
  TMS99XX_refresh;
  actualiza_video;
  sincronizar;
  evaluar_entradas;
  end;
end;

function coleco_getbyte(direccion:word): Byte; pascal;
begin
if ((direccion>$5fff) and (direccion<$8000)) then coleco_getbyte:=memoria[$6000+(direccion and $3ff)]
  else coleco_getbyte:=memoria[direccion];
end;

procedure coleco_putbyte(direccion:word;valor:byte);pascal;
begin
//Solo tiene $400 bytes de memoria RAM, hace mirror desde $6000 hasta la $7fff
if ((direccion>$5fff) and (direccion<$8000)) then memoria[$6000+(direccion and $3ff)]:=valor;
end;

function coleco_inbyte(puerto:word):byte;pascal;
var
  nPAux: Integer;
  nResult: Byte;
begin
  nResult:=$FF;
  case (puerto and $E0) of
    $E0: begin
      nJoyState[0]:=$FFFF;
      if teclado[SDLK_LEFT] then nJoyState[0]:=nJoyState[0] and $F7FF;
      if teclado[SDLK_RIGHT] then nJoyState[0]:=nJoyState[0] and $FDFF;
      if teclado[SDLK_UP] then nJoyState[0]:=nJoyState[0] and $FEFF;
      if teclado[SDLK_DOWN] then nJoyState[0]:=nJoyState[0] and $FBFF;
      if teclado[SDLK_SPACE] then nJoyState[0]:=nJoyState[0] and $FFBF;
      if teclado[SDLK_LCTRL] then nJoyState[0]:=nJoyState[0] and $FFBF;
      if teclado[SDLK_LALT] then nJoyState[0]:=nJoyState[0] and $BFFF;
      if teclado[SDLK_RETURN] then nJoyState[0]:=nJoyState[0] and $FFBF;
      if teclado[SDLK_0] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $0;
      if teclado[SDLK_1] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $1;
      if teclado[SDLK_2] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $2;
      if teclado[SDLK_3] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $3;
      if teclado[SDLK_4] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $4;
      if teclado[SDLK_5] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $5;
      if teclado[SDLK_6] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $6;
      if teclado[SDLK_7] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $7;
      if teclado[SDLK_8] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $8;
      if teclado[SDLK_9] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $9;
      if teclado[SDLK_Q] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $a;
      if teclado[SDLK_W] then nJoyState[0]:=(nJoyState[0] and $FFF0) or $b;
      nPAux:=(puerto shr 1) and $01;
      //Para emular los dos joysticks con las mismas teclas de funcion
      //Pero no las direcciones
      nJoyState[1]:=$fff0 or nJoyState[0] and $f;
      if nJoyMode then nPAux:=nJoyState[nPAux] shr 8
        else nPAux:=(nJoyState[nPAux] and $F0) or keycodes[nJoyState[nPAux] and $0F];
      nResult:=(nPAux or $B0) and $7F;
    end;
    $A0:if (puerto and $01)<>0 then nResult:=TMS99XX_register_r
          else nResult:= TMS99XX_vram_r;
  end;
  coleco_inbyte:=nResult;
end;

procedure coleco_outbyte(valor:byte;puerto:word);pascal;
begin
  case (puerto and $E0) of
    $80:nJoyMode:=false;
    $C0:nJoyMode:=true;
    $E0:SN76496Write(0,valor);
    $A0:if (puerto and $01)<>0 then TMS99XX_register_w(valor)
          else TMS99XX_vram_w(valor);
  end;
end;

procedure coleco_despues_instruccion;pascal;
begin
if hay_sonido then begin
  tsample[0][posicion_sonido]:=SN76496Update(0);
  inc(posicion_sonido);
  if posicion_sonido=long_sample then play_sonido;
end;
end;

function coleco_cargar_snapshot(data:pbyte;long:dword):boolean;pascal;
var
  f:byte;
  cadena:string;
  temp,temp2:integer;
  longitud,lbloque:integer;
  puntero,puntero2:pbyte;
begin
for f:=0 to 3 do begin
  cadena:=cadena+chr(data^);
  inc(data);
end;
if cadena<>'CLSN' then exit;
lbloque:=sacar_word(data);
inc(data,2);
temp:=data^;
inc(data);
temp2:=sacar_word(data);
inc(data,2);
temp2:=data^*65536+temp2;
inc(data);
getmem(puntero2,139400);
puntero:=puntero2;
ZDecompress(data,temp2,pointer(puntero),temp);
longitud:=0;
while longitud<>temp do begin
  cadena:='';
  for f:=0 to 3 do begin
        cadena:=cadena+chr(puntero^);
        inc(puntero);inc(longitud);
  end;
  lbloque:=sacar_word(puntero);
  inc(puntero,2);inc(longitud,2);
  if cadena='CRAM' then begin
    copymemory(@memoria[$2000],puntero,lbloque);
    inc(puntero,lbloque);inc(longitud,lbloque);
  end;
  if cadena='Z80R' then begin
    copymemory(main_z80,puntero,lbloque);
    inc(puntero,lbloque);inc(longitud,lbloque);
  end;
  if cadena='TMSR' then begin
    lbloque:=puntero^*65536+lbloque;
    inc(puntero);inc(longitud);
    copymemory(TMS,puntero,lbloque);
    inc(puntero,lbloque);inc(longitud,lbloque);
  end;
end;
freemem(puntero2);
end;

function abrir_cartucho(datos:pbyte;longitud:integer):boolean;
var
  ptemp:pbyte;
  crc:cardinal;
begin
abrir_cartucho:=false;
ptemp:=datos;
inc(ptemp,1);
if not(((datos^=$55) and (ptemp^=$aa)) or ((datos^=$aa) and (ptemp^=$55)) or ((datos^=$66) and (ptemp^=$99))) then exit;
zeromemory(@memoria[$2000],$E000);
copymemory(@memoria[$8000],datos,longitud);
reset_coleco;
//Tengo que parchear cuando se carga algun juego de atari...
//Hay un problema con las NMI's...
crc:=calccrc32(datos,longitud);
//En orden: pacman,centipede,defender,defender(alt),digdug,galaxian,galaxian(alt),joust,jungle hunt,matt patrol
tms.atari_patch:=(crc=$a40a07e8) or (crc=$17edbfd4) or (crc=$ffd16997) or (crc=$6cf594e5)
                    or (crc=$1038e0a1) or (crc=$fe1d0b6a) or (crc=$987bc7ac) or (crc=$52228000) or (crc=$e8858484)
                    or (crc=$90007079);
abrir_cartucho:=true;
end;

procedure abrir_coleco;pascal;
var
  fichero:file of byte;
  nombre,RomFile:string;
  datos:pbyte;
  ArchiveItem: TZFArchiveItem;
  zip_obj:tzipforge;
  longitud:integer;
begin
  form1.Pausa1Click(nil);
  RomFile:=Form1.OpenRom(StColecovision);
  nombre:=extension_fichero(RomFile);
  if nombre='ZIP' then begin
    zip_obj:=tzipforge.Create(nil);
    zip_obj.CleanupInstance;
    zip_obj.FileName:=RomFile;
    zip_obj.OpenArchive;
    if not(zip_obj.FindFirst('*.col',archiveitem)) then
      if not(zip_obj.FindFirst('*.rom',archiveitem)) then
        if not(zip_obj.FindFirst('*.csn',archiveitem)) then exit;
    nombre:=archiveitem.FileName;
    longitud:=archiveitem.UncompressedSize;
    getmem(datos,longitud);
    zip_obj.ExtractTobuffer(archiveitem.FileName,datos^,archiveitem.UncompressedSize);
    zip_obj.CloseArchive;
    zip_obj.Destroy;
  end else begin
    nombre:=extractfilename(RomFile);
    assignfile(fichero, RomFile);
    reset(fichero);
    longitud:=filesize(fichero);
    getmem(datos,longitud);
    blockread(fichero,datos^,longitud);
    closefile(fichero);
  end;
reset_coleco;
romfile:=extension_fichero(nombre);
if romfile='CSN' then coleco_cargar_snapshot(datos,longitud);
if ((romfile='COL') or (romfile='ROM')) then abrir_cartucho(datos,longitud);
child.Caption:=llamadas_maquina.caption+' - '+nombre;
//Restauro la llamada de interrupcion del TMS
tms.IRQ_Handler:=coleco_interrupt;
end;

procedure coleco_interrupt(int:boolean);pascal;
begin
  main_z80.pedir_nmi:=int;
end;

procedure coleco_grabar_snapshot;
var
  cantidad,long_final,datos:integer;
  fichero:file of byte;
  buffer:array[0..9] of byte;
  nombre:string;
  puntero,data,puntero2:pbyte;
begin
form1.savedialog1.InitialDir:=Directory.coleco_snap;
form1.saveDialog1.Filter := 'CSN Format (*.CSN)|*.CSN';
if form1.savedialog1.execute then begin
        nombre:=changefileext(form1.savedialog1.FileName,'.CSN');
        if FileExists(nombre) then begin
            datos:=MessageDlg(leng[idioma].mensajes[3], mtWarning, [mbYes]+[mbNo],0);
            if datos=IDNO then exit;
        end;
end else exit;
long_final:=0;
getmem(puntero,139400);
data:=puntero;
//Coleco RAM (longitud=$e000)
buffer[0]:=ord('C');
buffer[1]:=ord('R');
buffer[2]:=ord('A');
buffer[3]:=ord('M');
buffer[4]:=$e000 mod 256;
buffer[5]:=$e000 div 256;
copymemory(data,@buffer[0],6);
inc(data,6);inc(long_final,6);
copymemory(data,@memoria[$2000],$e000);
inc(data,$e000);inc(long_final,$e000);
//TMS9918 (longitud=131112)
cantidad:=sizeof(TTMS99XX);
buffer[0]:=ord('T');
buffer[1]:=ord('M');
buffer[2]:=ord('S');
buffer[3]:=ord('R');
buffer[4]:=(cantidad mod 65536) mod 256;
buffer[5]:=(cantidad mod 65536) div 256;
buffer[6]:=cantidad div 65536;
copymemory(data,@buffer[0],7);
inc(data,7);inc(long_final,7);
copymemory(data,TMS,cantidad);
inc(data,cantidad);inc(long_final,cantidad);
//Registros Z80 (longitud=64)
buffer[0]:=ord('Z');
buffer[1]:=ord('8');
buffer[2]:=ord('0');
buffer[3]:=ord('R');
buffer[4]:=64;
buffer[5]:=0;
copymemory(data,@buffer[0],6);
inc(data,6);inc(long_final,6);
copymemory(data,main_z80,64);
inc(data,64);inc(long_final,64);
getmem(puntero2,long_final);
ZCompress(puntero,long_final,pointer(puntero2),cantidad,zcMAX);
//Cabecera
buffer[0]:=ord('C'); //Nombre Bloque
buffer[1]:=ord('L');
buffer[2]:=ord('S');
buffer[3]:=ord('N');
buffer[4]:=1; //Longitud bloque (word)
buffer[5]:=0;
buffer[6]:=1; //Version Snapshot
buffer[7]:=(cantidad mod 65536) mod 256;
buffer[8]:=(cantidad mod 65536) div 256;
buffer[9]:=cantidad div 65536;
{$I-}
assignfile(fichero,nombre);
rewrite(fichero);
blockwrite(fichero,buffer[0],10,datos);
blockwrite(fichero,puntero2^,cantidad,datos);
{$I+}
close(fichero);
freemem(puntero);
freemem(puntero2);
end;

end.
