unit nes_ppu;

interface

uses constantes, windows;

procedure ppu_dma_spr(direccion:byte);
procedure reset_ppu;
procedure ppu_linea;
procedure putsprites(posicion:boolean);
procedure ppu_write(direccion:word;valor:byte);
procedure convertir_chars;
procedure pantalla_nes;

var
  ppu_control1,ppu_control2,ppu_status,sprite_ram_pos:byte;
  ppu_adress:parejas;
  sprite_ram:array[0..255] of byte;
  chars:array[0..1,0..255,0..63] of byte;
  pos_bg,pos_spt:byte;
  scroll_x,scroll_y,scroll_x2:byte;
  PPU    : array[0..3] of word;
//  PPU name_table0,name_table1,name_table2,name_table3:word;
  ppu_firstread,ppu_scroll_first,ppu_dir_first:boolean;
  cambios_pant,cambios_pant_ver,cambios_pant_hor:array [0..959] of boolean;
  nes_datos_pantalla:array[0..$3BFFF] of word;
  mhorizontal,mvertical:boolean;
  color_trans:byte;

implementation

uses nes;

procedure putchar(x,y,num_char:byte;table_name:word;despl:dword);
var
        f,g:byte;
        base_atrib,atrib_t:word;
        atrib:byte;
        pos:dword;
begin
//hay una paleta de 64 colores de los cuales se seleccionan
//16 para el fondo en la direccion $3f00 y otros 16 colores
//para los sprites en la direccion $3f10
//atributos en $3c0 con el siguiente sistema, hay 64 bytes que
//dividen la pantalla en cuadrados, cada cuadrado aporta 2 bits
//de la paleta usando los 8 bits para sub dividir la pantalla en
//4 cuadrados mas.
base_atrib:=table_name+$3c0;
atrib_t:=((x shr 2)+8*(y shr 2));  //calculo la posicion del atributo dentro de los 64bytes
case ((x and $3)+4*(y and $3)) of
    0,1,4,5:atrib:=(vrom[base_atrib+atrib_t] and 3) shl 2;
    2,3,6,7:atrib:=(vrom[base_atrib+atrib_t] and $0c);
    8,9,12,13:atrib:=(vrom[base_atrib+atrib_t] and $30) shr 2;
    10,11,14,15:atrib:=(vrom[base_atrib+atrib_t] and $c0) shr 4;
end;
pos:=despl+(x*8)+(y*8*512);
for g:=0 to 7 do
   for f:=0 to 7 do nes_datos_pantalla[pos+f+(g*512)]:=vrom[$3f00+chars[pos_bg][num_char][g+(f*8)]+atrib];
end;

procedure pantalla_nes;
var
        x,y:byte;
        char:byte;
begin
for y:=0 to 29 do
   for x:=0 to 31 do begin
        if cambios_pant[x+y*32] then begin
          //putchar(x,y,68,name_table0,0);
          putchar(x,y,vrom[PPU[0]+x+y*32],PPU[0],0);
          if mvertical then putchar(x,y,vrom[PPU[2]+x+y*32],PPU[2],$1E000);
          if mhorizontal then putchar(x,y,vrom[PPU[1]+x+y*32],PPU[1],$100);
          inc(char);
          cambios_pant[x+y*32]:=false;
        end;
   end;
end;

procedure putsprites(posicion:boolean);
var
  f:byte;
  pos_x,pos_y,x,y:byte;
  num_char,atrib:byte;
  temp:pword;
  flipx,flipy,pos_sprite:boolean;
  trans:byte;
begin
for f:=63 downto 0 do begin
 pos_sprite:=(sprite_ram[(f*4)+2] and $20)=$20;
 if posicion then trans:=0 else trans:=$FF;
 if pos_sprite=posicion then begin
  pos_y:=sprite_ram[f*4];
  pos_x:=sprite_ram[(f*4)+3];
  num_char:=sprite_ram[(f*4)+1];
  atrib:=(sprite_ram[(f*4)+2] and $3) shl 2;
  flipx:=(sprite_ram[(f*4)+2] and $40)=$40;
  flipy:=(sprite_ram[(f*4)+2] and $80)=$80;
  if flipx then begin
    for y:=0 to 7 do begin
      temp:=punbuf;
      for x:=7 downto 0 do begin
        temp^:=vrom[$3f10+chars[pos_spt][num_char][x+(y*8)]+atrib];
        inc(temp);
      end;
      if flipy then putpixel_trans(0,7-y,8,punbuf,0)
        else putpixel_trans(0,y,8,punbuf,0);
     end;
  end else begin
    for y:=0 to 7 do begin
      temp:=punbuf;
      for x:=0 to 7 do begin
        temp^:=vrom[$3f10+chars[pos_spt][num_char][x+(y*8)]+atrib];
        inc(temp);
      end;
      if flipy then putpixel_trans(0,7-y,8,punbuf,0)
        else putpixel_trans(0,y,8,punbuf,0);
     end;
  end;
  actualiza_trozo(0,0,8,8,8,pos_x,pos_y,8,8,0);
 end;
end;
end;

procedure convertir_chars;
var
  temp:word;
  f,g,ctemp1,ctemp2:byte;
begin
temp:=0;
for f:=0 to 255 do begin
 for g:=0 to 7 do begin
    ctemp1:=vrom[temp+(f*8)];
    ctemp2:=vrom[temp+8+(f*8)];
    chars[0][f][g+(0*8)]:=(ctemp1 shr 7)+(ctemp2 shr 7)*2;
    chars[0][f][g+(1*8)]:=((ctemp1 and 64) shr 6)+((ctemp2 and 64) shr 6)*2;
    chars[0][f][g+(2*8)]:=((ctemp1 and 32) shr 5)+((ctemp2 and 32) shr 5)*2;
    chars[0][f][g+(3*8)]:=((ctemp1 and 16) shr 4)+((ctemp2 and 16) shr 4)*2;
    chars[0][f][g+(4*8)]:=((ctemp1 and 8) shr 3)+((ctemp2 and 8) shr 3)*2;
    chars[0][f][g+(5*8)]:=((ctemp1 and 4) shr 2)+((ctemp2 and 4) shr 2)*2;
    chars[0][f][g+(6*8)]:=((ctemp1 and 2) shr 1)+((ctemp2 and 2) shr 1)*2;
    chars[0][f][g+(7*8)]:=(ctemp1 and 1)+(ctemp2 and 1)*2;
    inc(temp);
 end;
end;
temp:=0;
for f:=0 to 255 do begin
 for g:=0 to 7 do begin
    ctemp1:=vrom[temp+$1000+(f*8)];
    ctemp2:=vrom[temp+8+$1000+(f*8)];
    chars[1][f][g+(0*8)]:=(ctemp1 shr 7)+(ctemp2 shr 7)*2;
    chars[1][f][g+(1*8)]:=((ctemp1 and 64) shr 6)+((ctemp2 and 64) shr 6)*2;
    chars[1][f][g+(2*8)]:=((ctemp1 and 32) shr 5)+((ctemp2 and 32) shr 5)*2;
    chars[1][f][g+(3*8)]:=((ctemp1 and 16) shr 4)+((ctemp2 and 16) shr 4)*2;
    chars[1][f][g+(4*8)]:=((ctemp1 and 8) shr 3)+((ctemp2 and 8) shr 3)*2;
    chars[1][f][g+(5*8)]:=((ctemp1 and 4) shr 2)+((ctemp2 and 4) shr 2)*2;
    chars[1][f][g+(6*8)]:=((ctemp1 and 2) shr 1)+((ctemp2 and 2) shr 1)*2;
    chars[1][f][g+(7*8)]:=(ctemp1 and 1)+(ctemp2 and 1)*2;
    inc(temp);
 end;
end;
end;

procedure ppu_write(direccion:word;valor:byte);
begin
direccion:=direccion and $3fff;
case direccion of
  $2000..$23bf:begin
                vrom[direccion]:=valor;
                cambios_pant[direccion-$2000]:=true;
               end;
  $3f00..$3f1f:begin
                case (direccion and $F) of
                  0,4,8,$C:begin
                            vrom[(direccion and $F)+$3f00]:=valor;
                            vrom[(direccion and $F)+$3f10]:=valor;
                           end;
                  else vrom[direccion]:=valor;
                 end;
                color_trans:=vrom[$3f00];
                fillmemory(@nes_datos_pantalla[0],$3bfff,color_trans);
                fillmemory(@cambios_pant[0],$3c0,1);
               end;
  else vrom[direccion]:=valor;
end;
end;

procedure ppu_linea;
begin
pantalla_nes;
if (ppu_control2 and $10)=$10 then begin
  Scroll_x:=(PPU_Adress.w And 31) * 8 + Scroll_x2;
  Scroll_y:=((PPU_Adress.w div 32) And 31) * 8 Or ((PPU_Adress.w div $1000) And 7);
  putpixel_trans(0,nes_linea,512,@nes_datos_pantalla[(scroll_y+nes_linea)*512],0);
end;
end;

procedure ppu_dma_spr(direccion:byte);
begin
copymemory(@sprite_ram[0],@memoria[$100*direccion],$100);
//dec(nes_6502.v_contar,512);
//dec(nes_6502.h_contar,512);
end;

procedure reset_ppu;
begin
fillmemory(@vrom[0],$3fff,$FF);
ppu_control1:=0;
ppu_control2:=0;
ppu_status:=0;
sprite_ram_pos:=0;
ppu_adress.w:=0;
ppu_firstread:=true;
ppu_scroll_first:=true;
fillmemory(@cambios_pant[0],$3c0,1);
end;

end.
