unit z80ctc;

interface
uses z80daisy,timers;

const
  INTERRUPT=$80;
  INTERRUPT_ON=$80;
  INTERRUPT_OFF=$00;

  MODE=$40;
  MODE_TIMER=$00;
  MODE_COUNTER=$40;

  PRESCALER=$20;
  PRESCALER_256=$20;
  PRESCALER_16=$00;

  EDGE=$10;
  EDGE_FALLING=$00;
  EDGE_RISING=$10;

  TRIGGER=$08;
  TRIGGER_AUTO=$00;
  TRIGGER_CLOCK=$08;

  CONSTANT=$04;
  CONSTANT_LOAD=$04;
  CONSTANT_NONE=$00;

  RESET=$02;
  RESET_CONTINUE=$00;
  RESET_ACTIVE=$02;

  CONTROL=$01;
  CONTROL_VECTOR=$00;
  CONTROL_WORD=$01;

// these extra bits help us keep things accurate */
  WAITING_FOR_TRIG=$100;

  NOTIMER_0=(1 shl 0);
  NOTIMER_1=(1 shl 1);
  NOTIMER_2=(1 shl 2);
  NOTIMER_3=(1 shl 3);


type
  tintr=procedure(state:byte);pascal;
  tzero=procedure(n,ch,data:byte);pascal;
  tipo_ctc_channel=record
  	zc:tzero;					// zero crossing callbacks */
  	notimer:byte;			// no timer masks */
  	mode:word;				// current mode */
  	tconst:word;				// time constant */
  	down:word;				// down counter (clock mode only) */
  	extclk:byte;				// current signal from the external clock */
  	//timer --> los temporizadores van del 7 al 14
  	int_state:byte;			// interrupt status (for daisy chain) */
  end;

  tipo_z80ctc=record
	  vector:byte;				// interrupt vector */
	  period16:single;			// 16/system clock */
	  period256:single;			// 256/system clock */
	  intr:tintr;	// interrupt callback */
	  channel:array[0..3] of tipo_ctc_channel;			// data for each channel */
  end;
  ptipo_z80ctc=^tipo_z80ctc;

var
  z80_ctc:array[0..1] of ptipo_z80ctc;

procedure z80ctc_add(num_daisy,num_ctc:byte;cpu_num:byte;baseclock:double;notimer:word;intr:tintr;zc0,zc1,zc2:tzero);
procedure z80ctc_close(n:byte);

function z80ctc_w(n,ch,data:byte):byte;
function z80ctc_r(n,ch:byte):byte;

function z80ctc_irq_state(n:byte):byte;pascal;
function z80ctc_irq_ack(n:byte):byte;pascal;
procedure z80ctc_irq_reti(n:byte);pascal;
procedure z80ctc_reset(n:byte);pascal;

procedure z80ctc_trg_w(n,ch,data:byte);pascal;
procedure temporizador_ctc_0;pascal;
procedure temporizador_ctc_1;pascal;
procedure temporizador_ctc_2;pascal;
procedure temporizador_ctc_3;pascal;

implementation

procedure interrupt_check(n:byte);
begin
	// if we have a callback, update it with the current state */
	if addr(z80_ctc[n].intr)<>nil then begin
    if (z80ctc_irq_state(n) and Z80_DAISY_INT)<>0 then z80_ctc[n].intr(1)
      else z80_ctc[n].intr(0);
  end;
end;

function z80ctc_w(n,ch,data:byte):byte;
var
  mode:byte;
  period:single;
begin
	ch:=ch and 3;
	// get the current mode */
	mode:=z80_ctc[n].channel[ch].mode;
	// if we're waiting for a time constant, this is it */
	if ((mode and CONSTANT)= CONSTANT_LOAD) then begin
		// set the time constant (0 -> 0x100) */
    if data<>0 then z80_ctc[n].channel[ch].tconst:=data
      else z80_ctc[n].channel[ch].tconst:=$100;
		// clear the internal mode -- we're no longer waiting */
		z80_ctc[n].channel[ch].mode:=z80_ctc[n].channel[ch].mode and not(CONSTANT);
		// also clear the reset, since the constant gets it going again */
		z80_ctc[n].channel[ch].mode:=z80_ctc[n].channel[ch].mode and not(RESET);
		// if we're in timer mode.... */
		if ((mode and MODE)=MODE_TIMER) then begin
			// if we're triggering on the time constant, reset the down counter now */
			if ((mode and TRIGGER)=TRIGGER_AUTO) then begin
				if (z80_ctc[n].channel[ch].notimer=0) then begin
          if ((mode and PRESCALER)=PRESCALER_16) then period:=z80_ctc[n].period16
            else period:=z80_ctc[n].period256;
            //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
					period:=period*z80_ctc[n].channel[ch].tconst;
          timer[7+(n*4)+ch].enabled:=true;
          timer[7+(n*4)+ch].tiempo:=period;
					//timer_adjust_periodic(channel->timer, period, ch, period);
				end else begin
					timer[7+(n*4)+ch].enabled:=false;
        end;
			end else // else set the bit indicating that we're waiting for the appropriate trigger */
				z80_ctc[n].channel[ch].mode:=z80_ctc[n].channel[ch].mode or WAITING_FOR_TRIG;
		end;
		// also set the down counter in case we're clocking externally */
		z80_ctc[n].channel[ch].down:=z80_ctc[n].channel[ch].tconst;
    timer[7+(n*4)+ch].enabled:=true;
		// all done here */
		exit;
	end;
	// if we're writing the interrupt vector, handle it specially */
	if (((data and CONTROL)=CONTROL_VECTOR) and (ch=0)) then begin
		z80_ctc[n].vector:= data and $f8;
		exit;
	end;
	// this must be a control word */
	if ((data and CONTROL)=CONTROL_WORD) then begin
		// set the new mode */
		z80_ctc[n].channel[ch].mode:=data;
		// if we're being reset, clear out any pending timers for this channel */
		if ((data and RESET)= RESET_ACTIVE) then begin
			timer[7+(n*4)+ch].enabled:=false;
			// note that we don't clear the interrupt state here! */
    end;
	end;
end;

function z80ctc_r(n,ch:byte):byte;
var
  period:single;
begin
	ch:=ch and 3;
	// if we're in counter mode, just return the count */
	if (((z80_ctc[n].channel[ch].mode and MODE)=MODE_COUNTER) or ((z80_ctc[n].channel[ch].mode and WAITING_FOR_TRIG)<>0)) then begin
		z80ctc_r:=z80_ctc[n].channel[ch].down;
    exit;
  end	else begin // else compute the down counter value */
    if ((z80_ctc[n].channel[ch].mode and PRESCALER)= PRESCALER_16) then period:=z80_ctc[n].period16
      else period:=z80_ctc[n].period256;
		if timer[7+(n*4)+ch].enabled then
      // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			z80ctc_r:=round((timer[7+(n*4)+ch].tactual*period)+1) and $ff
		else z80ctc_r:=0;
	end;
end;

function z80ctc_irq_state(n:byte):byte;
var
  ch,state:byte;
begin
	state:=0;
	// loop over all channels */
	for ch:=0 to 3 do begin
		// if we're servicing a request, don't indicate more interrupts */
		if (z80_ctc[n].channel[ch].int_state and Z80_DAISY_IEO)<>0 then begin
			state:=state or Z80_DAISY_IEO;
      break;
    end;
		state:=state or z80_ctc[n].channel[ch].int_state;
	end;
	z80ctc_irq_state:=state;
end;

function z80ctc_irq_ack(n:byte):byte;pascal;
var
  ch:byte;
begin
	// loop over all channels */
	for ch:=0 to 3 do begin
		// find the first channel with an interrupt requested */
		if (z80_ctc[n].channel[ch].int_state and Z80_DAISY_INT)<>0 then begin
			// clear interrupt, switch to the IEO state, and update the IRQs */
			z80_ctc[n].channel[ch].int_state:=Z80_DAISY_IEO;
			interrupt_check(n);
      z80ctc_irq_ack:=z80_ctc[n].vector+ch*2;
			exit;
		end;
	end;
	z80ctc_irq_ack:=z80_ctc[n].vector;
end;


procedure z80ctc_irq_reti(n:byte);pascal;
var
  ch:byte;
begin
	// loop over all channels */
	for ch:=0 to 3 do begin
		// find the first channel with an IEO pending */
		if (z80_ctc[n].channel[ch].int_state and Z80_DAISY_IEO)<>0 then begin
			// clear the IEO state and update the IRQs */
			z80_ctc[n].channel[ch].int_state:=z80_ctc[n].channel[ch].int_state and not(Z80_DAISY_IEO);
			interrupt_check(n);
      exit;
		end;
	end;
end;

procedure z80ctc_reset(n:byte);pascal;
var
  ch:byte;
begin
	for ch:=0 to 3 do begin
		z80_ctc[n].channel[ch].mode:= RESET_ACTIVE;
		z80_ctc[n].channel[ch].tconst:=$100;
    timer[7+(n*4)+ch].enabled:=false;
		z80_ctc[n].channel[ch].int_state:=0;
	end;
  interrupt_check(n);
end;

procedure z80ctc_close(n:byte);
begin
  z80_ctc[n].intr:=nil;
  freemem(z80_ctc[n]);
  z80_ctc[n]:=nil;
end;

procedure temporizador_ctc(n,ch:byte);
begin
	// down counter has reached zero - see if we should interrupt */
	if ((z80_ctc[n].channel[ch].mode and INTERRUPT)=INTERRUPT_ON) then begin
		z80_ctc[n].channel[ch].int_state:=z80_ctc[n].channel[ch].int_state or Z80_DAISY_INT;
		interrupt_check(n);
	end;
	// generate the clock pulse */
	if addr(z80_ctc[n].channel[ch].zc)<>nil then begin
		z80_ctc[n].channel[ch].zc(n,0,1);
		z80_ctc[n].channel[ch].zc(n,0,0);
	end;
	// reset the down counter */
	z80_ctc[n].channel[ch].down:=z80_ctc[n].channel[ch].tconst;
end;

procedure temporizador_ctc_0;pascal;
begin
  temporizador_ctc(timer[7].dev_asociado,0);
end;

procedure temporizador_ctc_1;pascal;
begin
  temporizador_ctc(timer[8].dev_asociado,0);
end;

procedure temporizador_ctc_2;pascal;
begin
  temporizador_ctc(timer[9].dev_asociado,0);
end;

procedure temporizador_ctc_3;pascal;
begin
  temporizador_ctc(timer[10].dev_asociado,0);
end;

procedure z80ctc_trg_w(n,ch,data:byte);pascal;
var
  period:single;
begin
	// normalize data */
	if data<>0 then data:=1;
	// see if the trigger value has changed */
	if (data<>z80_ctc[n].channel[ch].extclk) then begin
		z80_ctc[n].channel[ch].extclk:= data;
		// see if this is the active edge of the trigger */
		if ((((z80_ctc[n].channel[ch].mode and EDGE)=EDGE_RISING) and (data<>0)) or (((z80_ctc[n].channel[ch].mode and EDGE)=EDGE_FALLING) and (data=0))) then begin
			// if we're waiting for a trigger, start the timer */
			if (((z80_ctc[n].channel[ch].mode and WAITING_FOR_TRIG)<>0) and ((z80_ctc[n].channel[ch].mode and MODE)=MODE_TIMER)) then begin
				if (z80_ctc[n].channel[ch].notimer=0) then begin
					if ((mode and PRESCALER)=PRESCALER_16) then period:=z80_ctc[n].period16
            else period:=z80_ctc[n].period256;
					period:=period*z80_ctc[n].channel[ch].tconst;
          //!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
					//timer_adjust_periodic(channel->timer, period, ch, period);
				end else begin
					timer[7+(n*4)+ch].enabled:=false;
				end;
			end;
			// we're no longer waiting */
			z80_ctc[n].channel[ch].mode:=z80_ctc[n].channel[ch].mode and not(WAITING_FOR_TRIG);
			// if we're clocking externally, decrement the count */
			if ((z80_ctc[n].channel[ch].mode and MODE)=MODE_COUNTER) then begin
				z80_ctc[n].channel[ch].down:=z80_ctc[n].channel[ch].down-1;
				// if we hit zero, do the same thing as for a timer interrupt */
				if (z80_ctc[n].channel[ch].down=0) then begin
					temporizador_ctc(n,ch);
				end;
			end;
		end;
	end;
end;

procedure z80ctc_add(num_daisy,num_ctc:byte;cpu_num:byte;baseclock:double;notimer:word;intr:tintr;zc0,zc1,zc2:tzero);
var
  ch:byte;
begin
z80daisy_llamadas[num_daisy].reset:=z80ctc_reset;
z80daisy_llamadas[num_daisy].state:=z80ctc_irq_state;
z80daisy_llamadas[num_daisy].ack:=z80ctc_irq_ack;
z80daisy_llamadas[num_daisy].reti:=z80ctc_irq_reti;
z80daisy_llamadas[num_daisy].device_num:=num_ctc;
getmem(z80_ctc[num_ctc],sizeof(tipo_z80ctc));
z80_ctc[num_ctc].period16:=baseclock/16;
z80_ctc[num_ctc].period256:=baseclock/256;
for ch:=0 to 3 do begin
		z80_ctc[num_ctc].channel[ch].notimer:=(notimer shr ch) and 1;
    //Cuando el contador de cada canal llegue a 0 llamar a esta funcion
    timer[7+num_ctc].ejecutar:=temporizador_ctc_0;
    timer[7+num_ctc].dev_asociado:=num_ctc;
    timer[7+num_ctc].cpu :=cpu_num;
    timer[8+num_ctc].ejecutar:=temporizador_ctc_1;
    timer[8+num_ctc].dev_asociado:=num_ctc;
    timer[8+num_ctc].cpu :=cpu_num;
    timer[9+num_ctc].ejecutar:=temporizador_ctc_2;
    timer[9+num_ctc].dev_asociado:=num_ctc;
    timer[9+num_ctc].cpu :=cpu_num;
    timer[10+num_ctc].ejecutar:=temporizador_ctc_3;
    timer[10+num_ctc].dev_asociado:=num_ctc;
    timer[10+num_ctc].cpu :=cpu_num;
end;
z80_ctc[num_ctc].intr:=intr;
z80_ctc[num_ctc].channel[0].zc:=zc0;
z80_ctc[num_ctc].channel[1].zc:=zc1;
z80_ctc[num_ctc].channel[2].zc:=zc2;
z80_ctc[num_ctc].channel[3].zc:=nil;
end;

end.
