unit z80pio;

interface
uses z80daisy;

const
  PIO_MODE0=00;		// output mode */
  PIO_MODE1=01;		// input  mode */
  PIO_MODE2=02;		// i/o    mode */
  PIO_MODE3=03;		// bit    mode */
// pio controll port operation (bit 0-3)
  PIO_OP_MODE=$0f;		// mode select
  PIO_OP_INTC=$07;		// interrupt controll
  PIO_OP_INTE=$03;		// interrupt enable

  // pio interrupt controll nit */
  PIO_INT_ENABLE=$80;		// ENABLE : 0=disable , 1=enable */
  PIO_INT_AND=$40;		// LOGIC  : 0=OR      , 1=AND    */
  PIO_INT_HIGH=$20;		// LEVEL  : 0=low     , 1=high   */
  PIO_INT_MASK=$10;		// MASK   : 0=off     , 1=on     */



type
  tintr_rdyr=procedure(state:byte);pascal;
  tportr=function:byte;pascal;
  tportw=procedure(data:byte);pascal;
  tipo_z80pio=record
	  vector:array[0..1] of byte;  // interrupt vector
	  intr:tintr_rdyr;            // Llamada a la interrupcion
	  portr:array[0..1] of tportr;          // Llamada cuando read
    portw:array[0..1] of tportw;                              //Llamada cuando lee
    rdyr:array[0..1] of tintr_rdyr;          // Llamada cuando RDY active
	  mode:array[0..1] of byte;                        // mode 00=in,01=out,02=i/o,03=bit*/
	  enable:array[0..1] of byte;                      // interrupt enable               */
	  mask:array[0..1] of byte;                        // mask folowers                  */
	  dir:array[0..1] of byte;                         // direction (bit mode)           */
	  rdy:array[0..1] of byte;                         // ready pin level                */
	  in_:array[0..1] of byte;                          // input port data                */
	  out_:array[0..1] of byte;                         // output port                    */
	  strobe:array[0..1] of byte;						// strobe inputs */
	  int_state:array[0..1] of byte; // interrupt status (daisy chain) */
  end;
  ptipo_z80pio=^tipo_z80pio;

var
  z80_pio:array[0..1] of ptipo_z80pio;

procedure z80pio_add(num_daisy,num_pio:byte;intr:tintr_rdyr;portr_a,portr_b:tportr;portw_a,portw_b:tportw;rdyr_a,rdyr_b:tintr_rdyr);
procedure z80pio_reset(n:byte);pascal;
procedure z80pio_close(n:byte);

function z80pio_r(n,offset:byte):byte;
procedure z80pio_w(n,offset,data:byte);

procedure z80pio_c_w(n,ch,data:byte);
function z80pio_c_r(n,ch:byte):byte;
procedure z80pio_d_w(n,ch,data:byte);
function z80pio_d_r(n,ch:byte):byte;
function z80pio_p_r(n,ch:byte):byte;
procedure z80pio_p_w(n,ch,data:byte);

function z80pio_irq_state(n:byte):byte;pascal;
function z80pio_irq_ack(n:byte):byte;pascal;
procedure z80pio_irq_reti(n:byte);pascal;

implementation

procedure set_rdy(n,ch,state:byte);
begin
  z80_pio[n].rdy[ch]:=state;
  if addr(z80_pio[n].rdyr[ch])<>nil then z80_pio[n].rdyr[ch](state);
end;

procedure interrupt_check(n:byte);
begin
	// if we have a callback, update it with the current state */
	if addr(z80_pio[n].intr)<>nil then	begin
    if (z80pio_irq_state(n) and Z80_DAISY_INT)<>0 then z80_pio[n].intr(1)
      else z80_pio[n].intr(0);
  end;
end;

procedure update_irq_state(n,ch:byte);
var
  old_state,data:byte;
  irq:boolean;
begin
	old_state:=z80_pio[n].int_state[ch];
	irq:=false;
	if ((z80_pio[n].mode[ch]=$13) or ((z80_pio[n].enable[ch] and PIO_INT_MASK)<>0)) then exit;
	// only check if interrupts are enabled */
	if (z80_pio[n].enable[ch] and PIO_INT_ENABLE)<>0 then begin
		// in mode 3, interrupts are tricky */
		if (z80_pio[n].mode[ch]=PIO_MODE3) then begin
			// fetch input data (ignore output lines) */
			data:=z80_pio[n].in_[ch] and z80_pio[n].dir[ch];
			// keep only relevant bits */
			data:=data and not(z80_pio[n].mask[ch]);
			// if active low, invert the bits */
			if (not(z80_pio[n].enable[ch] and PIO_INT_HIGH))<>0 then
				data:=data xor z80_pio[n].mask[ch];
			// if AND logic, interrupt if all bits are set */
			if (z80_pio[n].enable[ch] and PIO_INT_AND)<>0 then irq:=(data=z80_pio[n].mask[ch])
        // otherwise, interrupt if at least one bit is set */
      else irq:=(data<>0);
			// if portB, portA mode 2 check */
			if ((ch<>0) and (z80_pio[n].mode[0]=PIO_MODE2)) then begin
				if (z80_pio[n].rdy[ch]=0) then irq:=true;
      end;
    end	else // otherwise, just interrupt when ready is cleared */
			irq:=(z80_pio[n].rdy[ch]=0);
	end;
	if irq then	z80_pio[n].int_state[ch]:=z80_pio[n].int_state[ch] or Z80_DAISY_INT
	  else z80_pio[n].int_state[ch]:=z80_pio[n].int_state[ch] and not(Z80_DAISY_INT);
	if (old_state<>z80_pio[n].int_state[ch]) then interrupt_check(n);
end;

procedure z80pio_reset(n:byte);
var
  f:byte;
begin
  for f:=0 to 1 do begin
    z80_pio[n].mask[f]:=$ff;	// mask all on */
		z80_pio[n].enable[f]:=$00;	// disable     */
		z80_pio[n].mode[f]:=$01;	// mode input  */
		z80_pio[n].dir[f]:=$01;	// dir  input  */
		set_rdy(n,f,0);	// RDY = low   */
		z80_pio[n].out_[f]:=$00;	// outdata = 0 */
		z80_pio[n].int_state[f]:= 0;
		z80_pio[n].strobe[f]:= 0;
  end;
  interrupt_check(n);
end;

procedure z80pio_close(n:byte);
begin
  z80_pio[n].intr:=nil;
  z80_pio[n].portr[0]:=nil;
  z80_pio[n].portr[1]:=nil;
  z80_pio[n].portw[0]:=nil;
  z80_pio[n].portw[1]:=nil;
  z80_pio[n].rdyr[0]:=nil;
  z80_pio[n].rdyr[1]:=nil;
  freemem(z80_pio[n]);
  z80_pio[n]:=nil;
end;

//Control RW
procedure z80pio_c_w(n,ch,data:byte);
begin
	// There are only 2 channels */
	ch:=ch and $01;

	// load direction phase ? */
	if (z80_pio[n].mode[ch]=$13) then begin
		z80_pio[n].dir[ch]:= data;
		z80_pio[n].mode[ch]:=$03;
		exit;
	end;

	// load mask folows phase ? */
	if (z80_pio[n].enable[ch] and PIO_INT_MASK)<>0 then begin
		// load mask folows */
		z80_pio[n].mask[ch]:= data;
		z80_pio[n].enable[ch]:=z80_pio[n].enable[ch] and not(PIO_INT_MASK);
		exit;
	end;

	case (data and $0f) of
		PIO_OP_MODE:begin	// mode select 0=out,1=in,2=i/o,3=bit */
			z80_pio[n].mode[ch]:= (data shr 6);
			if (z80_pio[n].mode[ch]=$03) then z80_pio[n].mode[ch]:=$13;
			exit;
    end;
		PIO_OP_INTC:begin		// interrupt control */
			z80_pio[n].enable[ch]:=data and $f0;
			z80_pio[n].mask[ch]:=0;
    end;
		PIO_OP_INTE:begin		// interrupt enable controll */
			z80_pio[n].enable[ch]:=z80_pio[n].enable[ch] and not(PIO_INT_ENABLE);
			z80_pio[n].enable[ch]:=z80_pio[n].enable[ch] or (data and PIO_INT_ENABLE);
    end;
		else begin
			if (data and 1)=0 then begin
				z80_pio[n].vector[ch]:= data;
			end;
    end;
	end;

	// interrupt check */
	update_irq_state(n,ch);

end;

function z80pio_c_r(n,ch:byte):byte;
begin
//Error,no se puede leer el byte de control!!!
z80pio_c_r:=0;
end;

//Datos RW
procedure z80pio_d_w(n,ch,data:byte);
begin
	ch:=ch and $01;

	z80_pio[n].out_[ch]:=data;	// latch out data */
	if (z80_pio[n].mode[ch]=PIO_MODE3) then
    //llamadas falta!!!
		//devcb_call_write8(&z80pio->port_write[offset], 0, data & ~z80pio->dir[offset]);
	else;
		//devcb_call_write8(&z80pio->port_write[offset], 0, data);

	case z80_pio[0].mode[ch] of
		PIO_MODE0,PIO_MODE2:begin			// mode 0 output -mode 2 i/o */
			set_rdy(n, ch, 1); // ready = H */
			update_irq_state(n,ch);
			exit;
    end;
		PIO_MODE1,PIO_MODE3:exit;			// mode 1 input -  mode 0 bit */
	end;

end;

function z80pio_d_r(n,ch:byte):byte;
begin
	ch:=ch and $01;
	case z80_pio[n].mode[ch] of
		PIO_MODE0:begin			// mode 0 output */
			z80pio_d_r:=z80_pio[n].out_[ch];
      exit;
    end;
		PIO_MODE1:begin			// mode 1 input */
			set_rdy(n,ch, 1);	// ready = H */
			if addr(z80_pio[n].portr[ch])<>nil then z80_pio[n].in_[ch]:=z80_pio[n].portr[ch];
			update_irq_state(n,ch);
			z80pio_d_r:=z80_pio[n].in_[ch];
      exit;
    end;
		PIO_MODE2:begin			// mode 2 i/o */
			set_rdy(n, 1, 1); // brdy = H */
			if addr(z80_pio[n].portr[ch])<>nil then z80_pio[n].in_[ch]:=z80_pio[n].portr[ch];
			update_irq_state(n,ch);
			z80pio_d_r:=z80_pio[n].in_[ch];
      exit;
    end;
		PIO_MODE3:begin			// mode 3 bit */
			if addr(z80_pio[n].portr[ch])<>nil then z80_pio[n].in_[ch]:=z80_pio[n].portr[ch];
			z80pio_d_r:=(z80_pio[n].in_[ch] and z80_pio[n].dir[ch]) or (z80_pio[n].out_[ch] and not(z80_pio[n].dir[ch]));
	  end;
  end;
	z80pio_d_r:=0;
end;


//Puertos RW
procedure z80pio_p_w(n,ch,data:byte);
begin
  ch:=ch and $1;
	z80_pio[n].in_[ch]:=data;
	case (z80_pio[n].mode[ch]) of
		PIO_MODE0:exit;
    PIO_MODE1:begin
			          set_rdy(n,ch,0);
			          update_irq_state(n,ch);
              end;
		PIO_MODE2:begin
                ch:=1;	// only port A handshake and IRQ is use portB */
			          set_rdy(n,ch,0);
			          update_irq_state(n,ch);
              end;
		PIO_MODE3:update_irq_state(n,ch); // irq check */
	end;
end;

function z80pio_p_r(n,ch:byte):byte;
begin
	// There are only 2 channels */
	ch:=ch and $01;

	case z80_pio[n].mode[ch] of
		PIO_MODE2,PIO_MODE0:begin		// port A only */
			set_rdy(n,ch, 0);
			update_irq_state(n, ch);
    end;
		PIO_MODE1:;
		PIO_MODE3:begin
			z80pio_p_r:=(z80_pio[n].in_[ch] and z80_pio[n].dir[ch]) or (z80_pio[n].out_[ch] and not(z80_pio[n].dir[ch]));
      exit;
    end;
	end;
	z80pio_p_r:=z80_pio[n].out_[ch];
end;

//funciones entrada/salida
function z80pio_r(n,offset:byte):byte;
begin
  if (offset and 2)<>0 then begin
    z80pio_r:=z80pio_c_r(n, offset and 1)
  end else begin
    z80pio_r:=z80pio_d_r(n, offset and 1);
  end;
end;

procedure z80pio_w(n,offset,data:byte);
begin
	if (offset and 2)<>0 then z80pio_c_w(n, offset and 1, data)
	  else z80pio_d_w(n, offset and 1, data);
end;

//Interfaz daisy chain
function z80pio_irq_state(n:byte):byte;pascal;
var
  state,ch:byte;
begin
	state:=0;
	// loop over all channels */
	for ch:=0 to 1 do begin
		// if we're servicing a request, don't indicate more interrupts */
		if (z80_pio[n].int_state[ch] and Z80_DAISY_IEO)<>0 then begin
			state:=state or Z80_DAISY_IEO;
			break;
		end;
		state:=state or z80_pio[n].int_state[ch];
	end;
	z80pio_irq_state:=state;
end;


function z80pio_irq_ack(n:byte):byte;pascal;
var
  ch:byte;
begin
	// loop over all channels */
	for ch:=0 to 1 do
		// find the first channel with an interrupt requested */
		if (z80_pio[n].int_state[ch] and Z80_DAISY_INT)<>0 then begin
			// clear interrupt, switch to the IEO state, and update the IRQs */
			z80_pio[n].int_state[ch]:=Z80_DAISY_IEO;
			interrupt_check(n);
			z80pio_irq_ack:=z80_pio[n].vector[ch];
      exit;
		end;
	z80pio_irq_ack:=z80_pio[n].vector[0];
end;

procedure z80pio_irq_reti(n:byte);pascal;
var
  ch:byte;
begin
	// loop over all channels */
	for ch:=0 to 1 do
		// find the first channel with an IEO pending */
		if (z80_pio[n].int_state[ch] and Z80_DAISY_IEO)<>0 then begin
			// clear the IEO state and update the IRQs */
			z80_pio[n].int_state[ch]:=z80_pio[n].int_state[ch] and not(Z80_DAISY_IEO);
			interrupt_check(n);
      exit;
		end;
end;

procedure z80pio_add(num_daisy,num_pio:byte;intr:tintr_rdyr;portr_a,portr_b:tportr;portw_a,portw_b:tportw;rdyr_a,rdyr_b:tintr_rdyr);
begin
z80daisy_llamadas[num_daisy].reset:=z80pio_reset;
z80daisy_llamadas[num_daisy].state:=z80pio_irq_state;
z80daisy_llamadas[num_daisy].ack:=z80pio_irq_ack;
z80daisy_llamadas[num_daisy].reti:=z80pio_irq_reti;
z80daisy_llamadas[num_daisy].device_num:=num_pio;
getmem(z80_pio[num_pio],sizeof(tipo_z80pio));
z80_pio[num_pio].intr:=intr;
z80_pio[num_pio].portr[0]:=portr_a;
z80_pio[num_pio].portr[1]:=portr_b;
z80_pio[num_pio].portw[0]:=portw_a;
z80_pio[num_pio].portw[1]:=portw_b;
z80_pio[num_pio].rdyr[0]:=rdyr_a;
z80_pio[num_pio].rdyr[1]:=rdyr_b;
end;

end.
