unit amstrad_cpc;

interface
uses z80,sdl,windows,constantes,entradas,ay_8910,sysutils,arcade_const,zipforge,timers,upd765,cargar_dsk,forms,dialogs;

const
  cpc_paleta:array[0..31] of dword=(
        $808080,$808080,$00FF80,$FFFF80,
        $000080,$FF0080,$008080,$FF8080,
        $FF0080,$FFFF80,$FFFF00,$FFFFFF,
        $FF0000,$FF00FF,$FF8000,$FF80FF,
        $000080,$00FF80,$00FF00,$00FFFF,
        $000000,$0000FF,$008000,$0080FF,
        $800080,$80FF80,$80FF80,$80FFFF,
        $800000,$8000FF,$808000,$8080FF);

  cpc6128_rom:tipo_roms=(n:'cpc6128.rom';l:$8000;p:0);
  cpc464_rom:tipo_roms=(n:'cpc464.rom';l:$8000;p:0);
  cpc646_rom:tipo_roms=(n:'cpc664.rom';l:$8000;p:0);
  ams_rom:tipo_roms=(n:'amsdos.rom';l:$4000;p:0);
  pantalla_largo=400;

var
    cpc_mem:array[0..10,0..$3fff] of byte;
    marco:array[0..3] of byte;
    cpc_z80:preg_z80;
    cpc_llamadas:llam_z80;
    curpen,cpcmode:byte;
    lowrom,highrom,crtcvsync:boolean;
    cpcpal:array[0..16] of byte;
    pa,pb,pc,con:byte;
    crtcreg:array[0..31] of byte;
    curcrtcreg,keyline:byte;
    keyl_val:array[0..9] of byte;
    psg_function,rom_selected:byte;
    galines,galc:byte;
    linea_crt,vc,sc:word;
    pixel_visible,borde_hi,cont_vsync:word;
    pintado:boolean;
    ajuste_vert:integer;

procedure Cargar_amstrad_CPC;pascal;
function cpc_getbyte(direccion:word): Byte; pascal;
procedure cpc_putbyte(direccion:word;valor:byte);pascal;
function cpc_inbyte(puerto:word):byte;pascal;
procedure cpc_outbyte(valor:byte;puerto:word);pascal;
function iniciar_cpc:boolean;pascal;
procedure cpc_main;pascal;
procedure cpc_cerrar;pascal;
procedure cpc_reset;pascal;
function cpc_porta_read:byte;pascal;
procedure amstrad_tapes;pascal;
procedure amstrad_despues_instruccion(estados_t:word);pascal;
procedure amstrad_audio;pascal;
procedure amstrad_raised_z80;pascal;
procedure amstrad_loaddisk;pascal;

implementation
uses principal,tap_tzx;

procedure Cargar_amstrad_CPC;pascal;
begin
form1.panel2.Visible:=true;
case tipo_maquina of
  3:begin
      llamadas_maquina.caption:='CPC464';
      form1.CPC1.Checked:=true;
    end;
  4:begin
      llamadas_maquina.caption:='CPC664';
      form1.CPC6641.Checked:=true;
    end;
  5:begin
      llamadas_maquina.caption:='CPC6128';
      form1.CPC61281.Checked:=true;
    end;
end;
llamadas_maquina.iniciar:=iniciar_cpc;
llamadas_maquina.bucle_general:=cpc_main;
llamadas_maquina.reset:=cpc_reset;
llamadas_maquina.fps_max:=50;
llamadas_maquina.velocidad_cpu:=4000000;
llamadas_maquina.canales_usados:=2;
llamadas_maquina.cerrar:=cpc_cerrar;
llamadas_maquina.cintas:=amstrad_tapes;
llamadas_maquina.cartuchos:=amstrad_loaddisk;
end;

function cpc_getbyte(direccion:word): Byte; pascal;
begin
case direccion of
  0..$3fff:if lowrom then cpc_getbyte:=cpc_mem[8,direccion and $3fff]
            else cpc_getbyte:=memoria[direccion];
  $4000..$bfff:cpc_getbyte:=memoria[direccion];
  $c000..$ffff:if highrom then cpc_getbyte:=cpc_mem[rom_selected,direccion and $3fff]
                else cpc_getbyte:=memoria[direccion];
end;
end;

procedure cpc_putbyte(direccion:word;valor:byte);pascal;
begin
memoria[direccion]:=valor;
end;

procedure writega(val:byte);
begin
  case (val shr 6) of
          $0:if (val and $10)=0 then curpen:=val and 15 //Select pen to change*/
                else curpen:=$10;
          $1:cpcpal[curpen]:=val and 31;    //Change pen colour*/
          $2:begin   //ROM banking and mode switch*/
                cpcmode:=val and 3;
                // Update ROM status*/
                lowrom:=(val and 4)=0;
                highrom:=(val and 8)=0;
                if (val and $10)<>0 then begin
                   galines:=0;
                   cpc_z80.pedir_irq:=false;
                end;
            end;
        end;
end;

procedure ram_banks(val:byte);
procedure cambia_marco(pos_marco,new_valor:byte);
begin
if marco[pos_marco]<>new_valor then begin
      copymemory(@cpc_mem[marco[pos_marco],0],@memoria[pos_marco*$4000],$4000);
      marco[pos_marco]:=new_valor;
      copymemory(@memoria[pos_marco*$4000],@cpc_mem[new_valor,0],$4000);
end;
end;
begin
if (val shr 6)=3 then begin
  if tipo_maquina<>3 then begin //C6128 RAM Bank
    cambia_marco(0,0);
    cambia_marco(2,2);
    cambia_marco(3,3);
    case (val and 7) of
      0:cambia_marco(1,1);
      1:begin
          cambia_marco(1,1);
          cambia_marco(3,7);
        end;
      2:begin
          cambia_marco(0,4);
          cambia_marco(1,5);
          cambia_marco(2,6);
          cambia_marco(3,7);
        end;
      3:begin
          cambia_marco(1,3);
          cambia_marco(3,7);
        end;
      4:cambia_marco(1,4);
      5:cambia_marco(1,5);
      6:cambia_marco(1,6);
      7:cambia_marco(1,7);
    end;
  end;
end;
end;

procedure update_psg;
begin
  case ((psg_function and $c0) shr 6) of
    0:;  //Inactive
  	1:pa:=AYRead(0);		// b6 = 1 ? : Read from selected PSG register and make the register data available to PPI Port A */
  	2:AYWrite(pa,0);
  	3:AYControl(pa,0); //* b6 and b7 = 1 ? : The register will now be selected and the user can read from or write to it.  The register will remain selected until another is chosen.*/
  end;
end;

{Control Word:

   $80  bit 7   - Mode set flag (1=active)
   $60  bit 6-5 - Group A Mode selection
                    00 - Mode 0
                    01 - Mode 1
                    1x - Mode 2
   $10  bit 4   - Port A direction (1=input 0=output)
    8   bit 3   - Port C upper direction (1=input 0=output)
    4   bit 2   - Group B Mode selection
                    0 - Mode 0
                    1 - Mode 1
    2   bit 1   - Port B direction (1=input 0=output)
    1   bit 0   - Port C lower direction (1=input 0=output)

        Port A and Port C upper are in group A, and Port B and Port C lower
        are in group B


    Mode 0: Basic Input/Output
        In Mode 0, each of the ports (A, B and C) operate as independent
        ports for whom direction can be set independently.

        Port C Usage In Mode 0:

            bits 7-4    Input/Output A (direction specified by ctrl bit 3)
            bits 3-0    Input/Output B (direction specified by ctrl bit 0)
}

function readpia(puerto:word):byte;
var
  ret:byte;
begin
  ret:=$ff;
  case ((puerto and $300) shr 8) of
    $00:if (con and $10)<>0 then begin //Port A is connected to the PSG*/
              update_psg;
              ret:=pa;
          end;
    $01:if (con and $2)<>0 then begin  //Port B holds `switch settings' and inputs*/
            if crtcvsync then ret:=pb or (ear shl 1) or 1
              else ret:=pb or (ear shl 1);
          end;
    $02:begin
          ret:=pc;
          //if (con and 8)<>0 then ret:=ret or (pc and $F0);
          //if (con and 1)<>0 then ret:=ret or (pc and $0f);
        end;
  end;
  readpia:=ret;
end;

procedure writepia(puerto:word;valor:byte);
begin
  case ((puerto and $300) shr 8) of
    $00:if (con and $10)=0 then begin // Port A - connected to PSG*/
            pa:=valor;
            update_psg;
          end;
    $02:begin
            if (con and 8)=0 then begin
              pc:=(pc and $0f) or (valor and $f0);
              Psg_Function:=valor;
              update_psg;
            end;
            if (con and 1)=0 then begin
              pc:=(pc and $f0) or (valor and $0f);
              keyline:=valor and $f;
            end;
          end;
    $03:if (valor and $80)<>0 then begin
            con:=valor;
        end else begin
          if (valor and 1)<>0 then con:=con or (1 shl ((valor shr 1) and 7))
            else con:=con and not(1 shl ((valor shr 1) and 7));
        end;
    end;
end;

procedure writecrtc(port:word;val:byte);
var
  linea_visible:word;
begin
  case ((port and $300) shr 8) of
    $00:curcrtcreg:=val and 31;
    $01:begin
          crtcreg[curcrtcreg]:=val;
          linea_visible:=crtcreg[6]*(crtcreg[9]+1)-1;
          //linea_total:=(crtcreg[4]+1)*(crtcreg[9]+1)+crtcreg[5];
          borde_hi:=((((crtcreg[4]+1)-crtcreg[7])*(crtcreg[9]+1))+crtcreg[5]);
          if (crtcreg[4]+1)<crtcreg[7] then borde_hi:=linea_crt;//crtcreg[7]*(crtcreg[9]+1);
          //pixel_total:=(crtcreg[0]+1)*8;
          pixel_visible:=crtcreg[1]*8;
          ajuste_vert:=borde_hi-((312-linea_visible) div 2);
        end;
  end;
end;

function readcrtc(port:word):byte;
begin
  case ((port and $300) shr 8) of
    $00:if ((curcrtcreg>11) and (curcrtcreg<18)) then readcrtc:=crtcreg[curcrtcreg]
      else readcrtc:=0;
    else readcrtc:=0;
  end;
end;

procedure cpc_outbyte(valor:byte;puerto:word);pascal;
begin
  if (puerto and $8000)=0 then begin
    if (puerto and $4000)<>0 then writega(valor);
    ram_banks(valor);
  end;
  if (puerto and $4000)=0 then writecrtc(puerto,valor);
  if (((puerto and $2000)=0) and (tipo_maquina<>3)) then begin
    case valor of
      0:rom_selected:=9;
      7:rom_selected:=10;
    end;
  end;
  if (puerto and $0800)=0 then writepia(puerto,valor);
  if (puerto and $0480)=0 then WriteFDCData(valor);
end;

function cpc_inbyte(puerto:word):byte;pascal;
begin
  cpc_inbyte:=$FF;
    if (puerto and $4000)=0 then cpc_inbyte:=readcrtc(puerto);
    if (puerto and $0800)=0 then cpc_inbyte:=readpia(puerto);
    if (puerto and $0480)=0 then begin
      case (puerto and $101) of
        $100:cpc_inbyte:=ReadFDCStatus;
        $101:cpc_inbyte:=ReadFDCData;
      end;
    end;
end;

function iniciar_cpc:boolean;pascal;
var
  f:byte;
  colores:tpaleta;
begin
//Lateral
form1.Panel2.Visible:=true;
//Cinta
form1.GroupBox1.Visible:=false;
//Botones Lateral
form1.BitBtn9.visible:=true;
form1.BitBtn11.visible:=true;
form1.BitBtn10.visible:=true;
form1.BitBtn10.Glyph:=nil;
form1.ImageList2.GetBitmap(3,form1.BitBtn10.Glyph);
form1.BitBtn9.enabled:=true;
form1.BitBtn11.enabled:=true;
if tipo_maquina<>3 then form1.BitBtn10.enabled:=true;
iniciar_cpc:=false;
iniciar_audio(1);
p_final[0].x:=pantalla_largo;
p_final[0].y:=312;
p_final[1].x:=pantalla_largo;
p_final[1].y:=312;
iniciar_video;
for f:=0 to 31 do begin
  colores[f].r:=cpc_paleta[f] shr 16;
  colores[f].g:=(cpc_paleta[f] shr 8) and $FF;
  colores[f].b:=cpc_paleta[f] and $FF;
end;
poner_paleta(colores,32);
getmem(cpc_z80,sizeof(reg_z80));
zeromemory(cpc_z80,sizeof(reg_z80));
cpc_z80.numero_cpu:=0;
cpc_z80.daisy:=false;
cpc_z80.tframes:=4000000/llamadas_maquina.fps_max/312;
cpc_llamadas.getbyte:=cpc_getbyte;
cpc_llamadas.putbyte:=cpc_putbyte;
cpc_llamadas.outbyte:=cpc_outbyte;
cpc_llamadas.inbyte:=cpc_inbyte;
cpc_llamadas.despues_instruccion:=amstrad_despues_instruccion;
cpc_llamadas.raised_z80:=amstrad_raised_z80;
ay8910_init(1000000,0,cpc_porta_read,cpc_porta_read,nil,nil);
init_timer(0,0,4000000/44100,amstrad_audio,true);
case tipo_maquina of
  3:if not(cargar_roms(@memoria_temp[0],@cpc464_rom,'\cpc464.zip',1)) then exit;
  4:begin
      if not(cargar_roms(@memoria_temp[0],@ams_rom,'\cpc664.zip',1)) then exit;
      copymemory(@cpc_mem[10,0],@memoria_temp[0],$4000);
      if not(cargar_roms(@memoria_temp[0],@cpc646_rom,'\cpc664.zip',1)) then exit;
  end;
  5:begin
      if not(cargar_roms(@memoria_temp[0],@ams_rom,'\cpc6128.zip',1)) then exit;
      copymemory(@cpc_mem[10,0],@memoria_temp[0],$4000);
      if not(cargar_roms(@memoria_temp[0],@cpc6128_rom,'\cpc6128.zip',1)) then exit;
  end;
end;
copymemory(@cpc_mem[8,0],@memoria_temp[0],$4000);
copymemory(@cpc_mem[9,0],@memoria_temp[$4000],$4000);
cpc_reset;
iniciar_cpc:=true;
end;

procedure eventos_cpc;
begin
if (evento.teclado or evento.joystick) then begin
  //Line 0 KEY_UP,KEY_RIGHT,KEY_DOWN,0,0,0,KEY_HOME,0
  //if teclado[SDLK_UP] then keyl_val[0]:=(keyl_val[0] and $fe) else keyl_val[0]:=(keyl_val[0] or 1);
  //if teclado[SDLK_RIGHT] then keyl_val[0]:=(keyl_val[0] and $fd) else keyl_val[0]:=(keyl_val[0] or 2);
  //if teclado[SDLK_DOWN] then keyl_val[0]:=(keyl_val[0] and $fb) else keyl_val[0]:=(keyl_val[0] or 4);
  if (teclado[SDLK_f6] and teclado[SDLK_RSHIFT])then keyl_val[0]:=(keyl_val[0] and $ef) else keyl_val[0]:=(keyl_val[0] or $10);
  if (teclado[SDLK_f3] and teclado[SDLK_RSHIFT])then keyl_val[0]:=(keyl_val[0] and $df) else keyl_val[0]:=(keyl_val[0] or $20);
  if teclado[SDLK_HOME] then keyl_val[0]:=(keyl_val[0] and $bf) else keyl_val[0]:=(keyl_val[0] or $40);
  //Line 1 KEY_LEFT,KEY_INSERT
  //if teclado[SDLK_LEFT] then keyl_val[1]:=(keyl_val[1] and $fe) else keyl_val[1]:=(keyl_val[1] or 1);
  if teclado[SDLK_INSERT] then keyl_val[1]:=(keyl_val[1] and $fd) else keyl_val[1]:=(keyl_val[1] or 2);
  if (teclado[SDLK_f7] and teclado[SDLK_RSHIFT])then keyl_val[1]:=(keyl_val[1] and $fb) else keyl_val[1]:=(keyl_val[1] or $4);
  if teclado[SDLK_f5] then begin
    if (teclado[SDLK_RSHIFT]) then begin
        keyl_val[1]:=(keyl_val[1] and $ef);
    end else begin
        clear_disk(0);
        child.Caption:=llamadas_maquina.caption;
      end;
  end else begin
        keyl_val[1]:=(keyl_val[1] or $10)
  end;
  if teclado[SDLK_f1] then begin
    if teclado[SDLK_RSHIFT] then begin
      keyl_val[1]:=(keyl_val[1] and $df);
    end else begin
      if cinta_tzx.cargada then begin
        if cinta_tzx.play_tape then Form1.fStopCinta(nil)
          else Form1.fPlayCinta(nil);
      end;
    end;
  end else begin
    keyl_val[1]:=(keyl_val[1] or $20);
  end;
  if (teclado[SDLK_f2] and teclado[SDLK_RSHIFT])then keyl_val[1]:=(keyl_val[1] and $bf) else keyl_val[1]:=(keyl_val[1] or $40);
  //Line 2 KEY_DEL,0,KEY_ENTER,0,0,KEY_RSHIFT,KEY_BSLASH,KEY_CTRL
  if teclado[SDLK_DELETE] then keyl_val[2]:=(keyl_val[2] and $fe) else keyl_val[2]:=(keyl_val[2] or 1);
  if teclado[SDLK_RETURN] then keyl_val[2]:=(keyl_val[2] and $fb) else keyl_val[2]:=(keyl_val[2] or 4);
  if (teclado[SDLK_f4] and teclado[SDLK_RSHIFT])then keyl_val[2]:=(keyl_val[2] and $ef) else keyl_val[2]:=(keyl_val[2] or $10);
  if teclado[SDLK_LSHIFT] then keyl_val[2]:=(keyl_val[2] and $df) else keyl_val[2]:=(keyl_val[2] or $20);
  if teclado[SDLK_BACKSLASH] then keyl_val[2]:=(keyl_val[2] and $bf) else keyl_val[2]:=(keyl_val[2] or $40);
  if teclado[SDLK_LCTRL] then keyl_val[2]:=(keyl_val[2] and $7f) else keyl_val[2]:=(keyl_val[2] or $80);
  //Line 3 KEY_F2,KEY_MINUS,KEY_F1,KEY_P,KEY_SCOLON,KEY_QUOTA,KEY_SLASH,KEY_DOT
  //if teclado[SDLK_f] then keyl_val[3]:=(keyl_val[3] and $fe) else keyl_val[3]:=(keyl_val[3] or 1);
  if teclado[SDLK_MINUS] then keyl_val[3]:=(keyl_val[3] and $fd) else keyl_val[3]:=(keyl_val[3] or 2);
  if teclado[SDLK_less] then keyl_val[3]:=(keyl_val[3] and $fb) else keyl_val[3]:=(keyl_val[3] or 4);
  if teclado[SDLK_p] then keyl_val[3]:=(keyl_val[3] and $f7) else keyl_val[3]:=(keyl_val[3] or 8);
  if teclado[SDLK_semicolon] then keyl_val[3]:=(keyl_val[3] and $ef) else keyl_val[3]:=(keyl_val[3] or $10);
  if teclado[SDLK_QUOTE] then keyl_val[3]:=(keyl_val[3] and $df) else keyl_val[3]:=(keyl_val[3] or $20);
  if teclado[SDLK_slash] then keyl_val[3]:=(keyl_val[3] and $bf) else keyl_val[3]:=(keyl_val[3] or $40);
  //punto if teclado[SDLK_less] then keyl_val[3]:=(keyl_val[3] and $7f) else keyl_val[3]:=(keyl_val[3] or $80);
  //Line 4
  if teclado[SDLK_0] then keyl_val[4]:=(keyl_val[4] and $fe) else keyl_val[4]:=(keyl_val[4] or 1);
  if teclado[SDLK_9] then keyl_val[4]:=(keyl_val[4] and $fd) else keyl_val[4]:=(keyl_val[4] or 2);
  if teclado[SDLK_o] then keyl_val[4]:=(keyl_val[4] and $fb) else keyl_val[4]:=(keyl_val[4] or 4);
  if teclado[SDLK_i] then keyl_val[4]:=(keyl_val[4] and $f7) else keyl_val[4]:=(keyl_val[4] or 8);
  if teclado[SDLK_l] then keyl_val[4]:=(keyl_val[4] and $ef) else keyl_val[4]:=(keyl_val[4] or $10);
  if teclado[SDLK_k] then keyl_val[4]:=(keyl_val[4] and $df) else keyl_val[4]:=(keyl_val[4] or $20);
  if teclado[SDLK_m] then keyl_val[4]:=(keyl_val[4] and $bf) else keyl_val[4]:=(keyl_val[4] or $40);
  if teclado[SDLK_comma] then keyl_val[4]:=(keyl_val[4] and $7f) else keyl_val[4]:=(keyl_val[4] or $80);
  //Line 5 KEY_8,KEY_7,KEY_U,KEY_Y,KEY_H,KEY_J,KEY_N,KEY_SPACE
  if teclado[SDLK_8] then keyl_val[5]:=(keyl_val[5] and $fe) else keyl_val[5]:=(keyl_val[5] or 1);
  if teclado[SDLK_7] then keyl_val[5]:=(keyl_val[5] and $fd) else keyl_val[5]:=(keyl_val[5] or 2);
  if teclado[SDLK_u] then keyl_val[5]:=(keyl_val[5] and $fb) else keyl_val[5]:=(keyl_val[5] or 4);
  if teclado[SDLK_y] then keyl_val[5]:=(keyl_val[5] and $f7) else keyl_val[5]:=(keyl_val[5] or 8);
  if teclado[SDLK_h] then keyl_val[5]:=(keyl_val[5] and $ef) else keyl_val[5]:=(keyl_val[5] or $10);
  if teclado[SDLK_j] then keyl_val[5]:=(keyl_val[5] and $df) else keyl_val[5]:=(keyl_val[5] or $20);
  if teclado[SDLK_n] then keyl_val[5]:=(keyl_val[5] and $bf) else keyl_val[5]:=(keyl_val[5] or $40);
  if teclado[SDLK_space] then keyl_val[5]:=(keyl_val[5] and $7f) else keyl_val[5]:=(keyl_val[5] or $80);
  //Line 6 KEY_6,KEY_5,KEY_R,KEY_T,KEY_G,KEY_F,KEY_B,KEY_V
  if teclado[SDLK_6] then keyl_val[6]:=(keyl_val[6] and $fe) else keyl_val[6]:=(keyl_val[6] or 1);
  if teclado[SDLK_5] then keyl_val[6]:=(keyl_val[6] and $fd) else keyl_val[6]:=(keyl_val[6] or 2);
  if teclado[SDLK_r] then keyl_val[6]:=(keyl_val[6] and $fb) else keyl_val[6]:=(keyl_val[6] or 4);
  if teclado[SDLK_t] then keyl_val[6]:=(keyl_val[6] and $f7) else keyl_val[6]:=(keyl_val[6] or 8);
  if teclado[SDLK_g] then keyl_val[6]:=(keyl_val[6] and $ef) else keyl_val[6]:=(keyl_val[6] or $10);
  if teclado[SDLK_f] then keyl_val[6]:=(keyl_val[6] and $df) else keyl_val[6]:=(keyl_val[6] or $20);
  if teclado[SDLK_b] then keyl_val[6]:=(keyl_val[6] and $bf) else keyl_val[6]:=(keyl_val[6] or $40);
  if teclado[SDLK_v] then keyl_val[6]:=(keyl_val[6] and $7f) else keyl_val[6]:=(keyl_val[6] or $80);
  //Line 7 KEY_4,KEY_3,KEY_E,KEY_W,KEY_S,KEY_D,KEY_C,KEY_X
  if teclado[SDLK_4] then keyl_val[7]:=(keyl_val[7] and $fe) else keyl_val[7]:=(keyl_val[7] or 1);
  if teclado[SDLK_3] then keyl_val[7]:=(keyl_val[7] and $fd) else keyl_val[7]:=(keyl_val[7] or 2);
  if teclado[SDLK_e] then keyl_val[7]:=(keyl_val[7] and $fb) else keyl_val[7]:=(keyl_val[7] or 4);
  if teclado[SDLK_w] then keyl_val[7]:=(keyl_val[7] and $f7) else keyl_val[7]:=(keyl_val[7] or 8);
  if teclado[SDLK_s] then keyl_val[7]:=(keyl_val[7] and $ef) else keyl_val[7]:=(keyl_val[7] or $10);
  if teclado[SDLK_d] then keyl_val[7]:=(keyl_val[7] and $df) else keyl_val[7]:=(keyl_val[7] or $20);
  if teclado[SDLK_c] then keyl_val[7]:=(keyl_val[7] and $bf) else keyl_val[7]:=(keyl_val[7] or $40);
  if teclado[SDLK_x] then keyl_val[7]:=(keyl_val[7] and $7f) else keyl_val[7]:=(keyl_val[7] or $80);
  //Line 8 KEY_1,KEY_2,KEY_ESC,KEY_Q,KEY_TAB,KEY_A,KEY_CAPS,KEY_Z
  if teclado[SDLK_1] then keyl_val[8]:=(keyl_val[8] and $fe) else keyl_val[8]:=(keyl_val[8] or 1);
  if teclado[SDLK_2] then keyl_val[8]:=(keyl_val[8] and $fd) else keyl_val[8]:=(keyl_val[8] or 2);
  if teclado[SDLK_escape] then keyl_val[8]:=(keyl_val[8] and $fb) else keyl_val[8]:=(keyl_val[8] or 4);
  if teclado[SDLK_q] then keyl_val[8]:=(keyl_val[8] and $f7) else keyl_val[8]:=(keyl_val[8] or 8);
  if teclado[SDLK_tab] then keyl_val[8]:=(keyl_val[8] and $ef) else keyl_val[8]:=(keyl_val[8] or $10);
  if teclado[SDLK_a] then keyl_val[8]:=(keyl_val[8] and $df) else keyl_val[8]:=(keyl_val[8] or $20);
  if teclado[SDLK_CAPSLOCK] then keyl_val[8]:=(keyl_val[8] and $bf) else keyl_val[8]:=(keyl_val[8] or $40);
  if teclado[SDLK_z] then keyl_val[8]:=(keyl_val[8] and $7f) else keyl_val[8]:=(keyl_val[8] or $80);
  //Line 9  JOY UP,JOY DOWN,JOY LEFT,JOY RIGHT,FIRE1,FIRE2, - ,KEY_BACKSP
  if teclado[SDLK_UP] then keyl_val[9]:=(keyl_val[9] and $fe) else keyl_val[9]:=(keyl_val[9] or 1);
  if teclado[SDLK_DOWN] then keyl_val[9]:=(keyl_val[9] and $fd) else keyl_val[9]:=(keyl_val[9] or 2);
  if teclado[SDLK_LEFT] then keyl_val[9]:=(keyl_val[9] and $fb) else keyl_val[9]:=(keyl_val[9] or 4);
  if teclado[SDLK_RIGHT] then keyl_val[9]:=(keyl_val[9] and $f7) else keyl_val[9]:=(keyl_val[9] or $8);
  if teclado[SDLK_RCTRL] then keyl_val[9]:=(keyl_val[9] and $ef) else keyl_val[9]:=(keyl_val[9] or $10);
  if teclado[SDLK_RALT] then keyl_val[9]:=(keyl_val[9] and $df) else keyl_val[9]:=(keyl_val[9] or $20);
  if teclado[SDLK_BACKSPACE] then keyl_val[9]:=(keyl_val[9] and $7f) else keyl_val[9]:=(keyl_val[9] or $80);
end;
end;

function cpc_porta_read:byte;pascal;
begin
  cpc_porta_read:=keyl_val[keyline];
end;

procedure amstrad_raised_z80;pascal;
begin
  galines:=galines and $1f;
end;

{Registros 6845
0 --> Numero de caracteres largo TOTAL (MENOS UNO)
      En el CPC 63
1 --> El ancho en caracteres VISIBLES de una fila.
2 --> Caracter en el que se produce el HSYNC
3 --> 4bits lo: Numero de caracteres que tarda el HSYNC
      4bits hi: Numero de caracteres que tarda el VSYNC
4 --> La altura en caracteres TOTAL de la pantalla
      En el CPC 38
5 --> Numero de lineas que faltan para completar la pantalla cuando termina un frame
      en el CPC 0
6 --> La altura en caracteres VISIBLES de la pantalla.
7 --> Numero de caracter en el que tiene que ocurrir el VSYNC
9 --> Numero de lineas por caracter (menos una), cuantas lineas necesito para dibujar un caracter
      Fijo a 7

Altura TOTAL= R4*(R9+1) = 39*8 = 312 lineas
Ancho TOTAL = (R0+1)*8 = 64*8 = 512 pixels
}

procedure actualiza_linea(linea:word);
var
 x,cy:integer;
 addr,addr2:dword;
 val,p1,p2,p3,p4,p5,p6,p7,p8:byte;
 ptemp:pword;
 ma,borde:word;
 temp:byte;
begin                       //MIRAR EL DEFENDER OF THE CROWN!!!!
cy:=linea-borde_hi;
if ((cy<0) or (linea>311)) then exit;
borde:=(pantalla_largo-pixel_visible) div 2;
linea_recta(0,linea,cpcpal[$10],borde,1);
ma:=(crtcreg[13] or (crtcreg[12] shl 8));
addr2:=((ma and $3000) shl 2) or ((ma and $03ff) shl 1);
addr:=addr2+((cy mod (crtcreg[9]+1)) shl 11)+((cy div (crtcreg[9]+1))*(crtcreg[1] shl 1));
x:=0;
while (x<pixel_visible) do begin
  val:=memoria[addr];
  inc(addr);
  case cpcmode of
    0:begin
        // 1 5 3 7    0 4 2 6
        p1:=((val and 2) shl 2) or ((val and $20) shr 3) or ((val and 8) shr 2) or ((val and $80) shr 7);
        p2:=((val and 1) shl 3) or ((val and $10) shr 2) or ((val and 4) shr 1) or ((val and $40) shr 6);
        ptemp:=punbuf;
        ptemp^:=paleta[cpcpal[p1]];
        inc(ptemp);
        ptemp^:=paleta[cpcpal[p1]];
        inc(ptemp);
        ptemp^:=paleta[cpcpal[p2]];
        inc(ptemp);
        ptemp^:=paleta[cpcpal[p2]];
        putpixel(x+borde,linea,4,punbuf,1);
    end;
    1:begin
        // 3 7      2 6      1 5      0 4
        p1:=((val and $80) shr 7)+((val and $8) shr 2);
        p2:=((val and $40) shr 6)+((val and $4) shr 1);
        p3:=((val and $20) shr 5)+((val and $2) shr 0);
        p4:=((val and $10) shr 4)+((val and 1) shl 1);
        ptemp:=punbuf;
        ptemp^:=paleta[cpcpal[p1]];
        inc(ptemp);
        ptemp^:=paleta[cpcpal[p2]];
        inc(ptemp);
        ptemp^:=paleta[cpcpal[p3]];
        inc(ptemp);
        ptemp^:=paleta[cpcpal[p4]];
        putpixel(x+borde,linea,4,punbuf,1);
      end;
    2:begin
        // 7 6 5 4 3 2 1 0
        p1:=((val and $80) shr 7);
        p2:=((val and $40) shr 6);
        p3:=((val and $20) shr 5);
        p4:=((val and $10) shr 4);
        p5:=((val and $8) shr 3);
        p6:=((val and $4) shr 2);
        p7:=((val and $2) shr 1);
        p8:=((val and $1) shr 0);
        ptemp:=punbuf;
        ptemp^:=paleta[cpcpal[p1]];
        inc(ptemp);
        ptemp^:=paleta[cpcpal[p3]];
        inc(ptemp);
        ptemp^:=paleta[cpcpal[p5]];
        inc(ptemp);
        ptemp^:=paleta[cpcpal[p7]];
        {inc(ptemp);
        ptemp^:=paleta[cpcpal[p5]];
        inc(ptemp);
        ptemp^:=paleta[cpcpal[p6]];
        inc(ptemp);
        ptemp^:=paleta[cpcpal[p7]];
        inc(ptemp);
        ptemp^:=paleta[cpcpal[p8]];}
        putpixel(x+borde,linea,4,punbuf,1);
    end;
  end;
  x:=x+4;
end;
linea_recta(pixel_visible+borde,linea,cpcpal[$10],borde,1);
end;

procedure ver_linea;
begin
if (linea_crt<(crtcreg[5] and 31)) then begin
    linea_crt:=linea_crt+1;
    galines:=galines+1;
    if (cont_vsync<>0) then begin
      cont_vsync:=cont_vsync-1;
      if (cont_vsync=0) then crtcvsync:=false;
    end;
    if (galc<>0) then begin
      galc:=galc-1;
      if (galc=0) then begin
        if (galines>=32) then cpc_z80.pedir_irq:=true;
        galines:=0;
      end;
    end;
    exit;
end;
if (vc<crtcreg[6]) then actualiza_linea(linea_crt)
  else linea_recta(0,linea_crt,cpcpal[$10],pantalla_largo,1);
if (sc=crtcreg[9]) then begin
  sc:=0;
  vc:=(vc+1) and 127;
end else //if (not(linea_crt<(crtcreg[5] and 31))) then
    begin
  sc:=(sc+1) and 31;
end;
linea_crt:=linea_crt+1;
galines:=galines+1;
if (galines=52) then begin
  galines:=0;
  cpc_z80.pedir_irq:=true;
end;
if (cont_vsync<>0) then begin
  cont_vsync:=cont_vsync-1;
  if (cont_vsync=0) then crtcvsync:=false;
end;
if (galc<>0) then begin
  galc:=galc-1;
  if (galc=0) then begin
    if (galines>=32) then cpc_z80.pedir_irq:=true;
    galines:=0;
  end;
end;
if (vc=crtcreg[4]+1) then begin
  sc:=0;
  vc:=0;
  actualiza_trozo(0,0,pantalla_largo,312,1,0,0,pantalla_largo,312,0);
  actualiza_video;
end;
if ((vc=crtcreg[7]) and (sc=0)) then begin
 cont_vsync:=crtcreg[3] shr 4;
 galc:=2;
 crtcvsync:=true;
 linea_crt:=0;
 actualiza_trozo(0,0,pantalla_largo,312,1,0,0,pantalla_largo,312,0);
 actualiza_video;
 eventos_cpc;
 sincronizar;
 evaluar_entradas;
 pintado:=true;
end;
if linea_crt=312 then begin
  linea_crt:=0;
  crtcvsync:=false;
  //actualiza_trozo(0,0,pantalla_largo,312,1,0,0,pantalla_largo,312,0);
  //actualiza_video;
  eventos_cpc;
  if not(pintado) then sincronizar;
  if pintado then begin
    dec(frames_sec);
    pintado:=false;
  end;
  evaluar_entradas;
end;
end;

procedure cpc_main;pascal;
var
  frame:single;
begin
iniciar_entradas;
//312 lineas 79872,2044 estados T por frame 256 estados T por linea
frame:=0;
while EmuStatus=EsRuning do begin
  frame:=frame+cpc_z80.tframes;
  run_z80(cpc_z80,cpc_llamadas,trunc(frame));
  frame:=frame-cpc_z80.contador;
  cpc_z80.contador:=0;
  ver_linea;
end;
end;

procedure cpc_cerrar;pascal;
begin
if cpc_z80<>nil then freemem(cpc_z80);
cpc_z80:=nil;
end;

procedure cpc_reset;pascal;
var
  f:byte;
begin
reset_audio;
if cinta_tzx.cargada then cinta_tzx.play_once:=false;
for f:=0 to 7 do zeromemory(@cpc_mem[f,0],$4000);
zeromemory(@memoria[0],$10000);
ear:=0;
reset_z80(cpc_z80);
cpc_z80.im2_lo:=$ff;
cpc_z80.tabla:=1;
ay8910_reset(0);
ResetFDC;
//Init GA
curpen:=0;
cpcmode:=1;
marco[0]:=0;
marco[1]:=1;
marco[2]:=2;
marco[3]:=3;
lowrom:=true;
highrom:=true;
rom_selected:=9;
//Init PIA
pa:=0;
pb:=$3e;
pc:=130;
//CRT
linea_crt:=0;
curcrtcreg:=0;
cont_vsync:=0;
crtcvsync:=false;
vc:=0;
sc:=0;
//Keyboard
for f:=0 to 9 do keyl_val[f]:=$FF;
end;

procedure amstrad_tapes;pascal;
var
  datos,temp:pbyte;
  file_size:integer;
  zip_obj:tzipforge;
  ArchiveItem: TZFArchiveItem;
  nombre,extension:string;
  fichero:file of byte;
  resultado:boolean;
  x,y:word;
begin
  form1.OpenDialog1.Filter:='CPC Tape (*.cdt;*.csw;*.wav;*.zip)|*.cdt;*.csw;*.wav;*.zip';
  form1.opendialog1.InitialDir:=directory.amstrad_tap;
  extension:=extension_fichero(form1.OpenDialog1.FileName);
  if form1.OpenDialog1.execute then begin
    extension:=extension_fichero(form1.OpenDialog1.FileName);
    if ((extension='CDT') or (extension='CSW') or (extension='WAV')) then begin
         {$I-}
         filemode:=fmOpenRead;
         nombre:=form1.OpenDialog1.FileName;
         assignfile(fichero,nombre);
         reset(fichero);
         if ioresult<>0 then exit;
         file_size:=filesize(fichero);
         getmem(datos,file_size);
         blockread(fichero,datos^,file_size);
         closefile(fichero);
         filemode:=fmOpenReadWrite;
         {$I+}
    end;
    if extension='ZIP' then begin
         zip_obj:=tzipforge.Create(nil);
         zip_obj.CleanupInstance;
         zip_obj.FileName:=form1.OpenDialog1.FileName;
         zip_obj.OpenArchive;
         zip_obj.FindFirst('*.cdt',archiveitem);
         nombre:=archiveitem.FileName;
         if nombre='' then zip_obj.FindFirst('*.csw',archiveitem);
         nombre:=archiveitem.FileName;
         if nombre='' then zip_obj.FindFirst('*.wav',archiveitem);
         nombre:=archiveitem.FileName;
         if nombre='' then exit;
         file_size:=archiveitem.UncompressedSize;
         getmem(datos,file_size);
         temp:=datos;
         zip_obj.ExtractTobuffer(archiveitem.StoredPath+archiveitem.FileName,temp^,file_size);
      end;
    extension:=extension_fichero(nombre);
    if extension='CDT' then resultado:=abrir_tzx(datos,file_size);
    if extension='CSW' then resultado:=abrir_csw(datos,file_size);
    if extension='WAV' then resultado:=abrir_wav(datos,file_size);
    if resultado then begin
        form1.edit1.Text:=nombre;
        form1.GroupBox1.Visible:=true;
        form1.BitBtn16.Enabled:=false;
        x:=dame_x_total;
        y:=dame_y_total;
        if cinta_tzx.cargada then begin
          form1.groupbox1.Left:=x+23;
          x:=x+350;
        end;
        x:=x+getosversion;
        y:=y+150;
        arregla_pos(x,y);
    end;
    freemem(datos);
    datos:=nil;
    form1.BitBtn15.Enabled:=true;
    cinta_tzx.play_tape:=false;
    directory.amstrad_tap:=extractfiledir(form1.OpenDialog1.filename);
  end;
end;

procedure amstrad_despues_instruccion(estados_t:word);pascal;
begin
if cinta_tzx.cargada then begin
  if cinta_tzx.play_tape then begin
        inc(estados_tzx,estados_t);
        play_cinta_tzx;
  end;
end;
end;

procedure amstrad_audio;pascal;
begin
if hay_sonido then begin
  tsample[0][posicion_sonido]:=ear*$20;//renderbyte(0)^;
  tsample[1][posicion_sonido]:=renderbyte(0)^;
  inc(posicion_sonido);
  if posicion_sonido=long_sample then play_sonido;
end;
end;

procedure amstrad_loaddisk;pascal;
begin
Application.createform(Tform6, form6);
try
  form6.showmodal;
finally
  form6.Release;
  FreeAndNil(form6);
end;
end;


end.
