unit dac;

interface
uses windows;

type
 dac_info=record
	  output:smallint;
	  UnsignedVolTable:array[0..256-1] of word;
	  SignedVolTable:array[0..256-1] of smallint;
  end;
  pdac_info=^dac_info;
var
  chip_dac:array[0..2] of pdac_info;

procedure dac_start(numero:byte);
function DAC_update(numero:byte):smallint;
procedure DAC_signed_data_w(numero,data:byte);
procedure DAC_data_w(numero,data:byte);
procedure dac_reset(numero:byte);
procedure dac_close(numero:byte);

implementation

function DAC_update(numero:byte):smallint;
var
  chip:pdac_info;
begin
  chip:=chip_dac[numero];
	DAC_update:=chip.output;
end;

procedure DAC_data_w(numero,data:byte);
var
  chip:pdac_info;
begin
  chip:=chip_dac[numero];
	chip.output:=chip.UnsignedVolTable[data];
end;

procedure DAC_signed_data_w(numero,data:byte);
var
  chip:pdac_info;
begin
  chip:=chip_dac[numero];
	chip.output:=chip.SignedVolTable[data];
end;

procedure DAC_build_voltable(numero:byte);
var
  chip:pdac_info;
  i:integer;
begin
  chip:=chip_dac[numero];
	// build volume table (linear) */
	for i:=0 to 255 do begin
		chip.UnsignedVolTable[i]:=i*$101 div 2;	// range      0..32767 */
		chip.SignedVolTable[i]:=i*$101 - $8000;	// range -32768..32767 */
	end;
end;

procedure dac_reset(numero:byte);
var
  chip:pdac_info;
begin
  chip:=chip_dac[numero];
  chip.output:=0;
end;

procedure dac_start(numero:byte);
var
  chip:pdac_info;
begin
  if chip_dac[numero]=nil then getmem(chip_dac[numero],sizeof(dac_info));
  chip:=chip_dac[numero];
	zeromemory(chip,sizeof(dac_info));
	DAC_build_voltable(numero);
  dac_reset(numero);
end;

procedure dac_close(numero:byte);
begin
  if chip_dac[numero]<>nil then begin
    freemem(chip_dac[numero]);
    chip_dac[numero]:=nil;
  end;
end;


end.
