unit vlm_5030;

interface
uses windows,constantes;

const
  FR_SIZE=4;
  // samples per interpolator */
  IP_SIZE_SLOWER=(240 div FR_SIZE);
  IP_SIZE_SLOW=(200 div FR_SIZE);
  IP_SIZE_NORMAL=(160 div FR_SIZE);
  IP_SIZE_FAST=(120 div FR_SIZE);
  IP_SIZE_FASTER=(80 div FR_SIZE);

	PH_RESET=0;
	PH_IDLE=1;
	PH_SETUP=2;
	PH_WAIT=3;
	PH_RUN=4;
	PH_STOP=5;
	PH_END=6;

// ROM Tables */
VLM5030_speed_table:array[0..8-1] of integer=(
 IP_SIZE_NORMAL,
 IP_SIZE_FAST,
 IP_SIZE_FASTER,
 IP_SIZE_FASTER,
 IP_SIZE_NORMAL,
 IP_SIZE_SLOWER,
 IP_SIZE_SLOW,
 IP_SIZE_SLOW
);

// This is the energy lookup table */
// sampled from real chip */
  energytable:array[0..$20-1] of byte=(
	  0,  2,  4,  6, 10, 12, 14, 18, // 0-7
	 22, 26, 30, 34, 38, 44, 48, 54, //  8-15
	 62, 68, 76, 84, 94,102,114,124, // 16-23
	136,150,164,178,196,214,232,254  // 24-31
  );
// This is the pitch lookup table */
  pitchtable:array[0..$20-1] of byte=(
   1,                               // 0     : random mode */
   22,                              // 1     : start=22    */
   23, 24, 25, 26, 27, 28, 29, 30,  //  2- 9 : 1step       */
   32, 34, 36, 38, 40, 42, 44, 46,  // 10-17 : 2step       */
   50, 54, 58, 62, 66, 70, 74, 78,  // 18-25 : 4step       */
   86, 94, 102,110,118,126          // 26-31 : 8step       */
   );

  K1_table:array[0..63] of integer= (
  -24898,  -25672,  -26446,  -27091,  -27736,  -28252,  -28768,  -29155,
  -29542,  -29929,  -30316,  -30574,  -30832,  -30961,  -31219,  -31348,
  -31606,  -31735,  -31864,  -31864,  -31993,  -32122,  -32122,  -32251,
  -32251,  -32380,  -32380,  -32380,  -32509,  -32509,  -32509,  -32509,
   24898,   23995,   22963,   21931,   20770,   19480,   18061,   16642,
   15093,   13416,   11610,    9804,    7998,    6063,    3999,    1935,
       0,   -1935,   -3999,   -6063,   -7998,   -9804,  -11610,  -13416,
  -15093,  -16642,  -18061,  -19480,  -20770,  -21931,  -22963,  -23995
);

  K2_table:array[0..31] of integer= (
       0,   -3096,   -6321,   -9417,  -12513,  -15351,  -18061,  -20770,
  -23092,  -25285,  -27220,  -28897,  -30187,  -31348,  -32122,  -32638,
       0,   32638,   32122,   31348,   30187,   28897,   27220,   25285,
   23092,   20770,   18061,   15351,   12513,    9417,    6321,    3096
);
  K3_table:array[0..15] of integer= (
       0,   -3999,   -8127,  -12255,  -16384,  -20383,  -24511,  -28639,
   32638,   28639,   24511,   20383,   16254,   12255,    8127,    3999
);
  K5_table:array[0..7] of integer= (
       0,   -8127,  -16384,  -24511,   32638,   24511,   16254,    8127
);


type

  vlm5030_info=record
	  rom:pbyte;
	  address_mask:integer;
	  address:word;
	  pin_BSY:byte;
	  pin_ST:byte;
	  pin_VCU:byte;
	  pin_RST:byte;
	  latch_data:byte;
	  vcu_addr_h:word;
	  parameter:byte;
	  phase:byte;
	  // state of option paramter */
	  frame_size:integer;
	  pitch_offset:integer;
	  interp_step:byte;

	  interp_count:byte;       // number of interp periods    */
	  sample_count:byte;       // sample number within interp */
	  pitch_count:byte;
	// these contain data describing the current and previous voice frames */
	  old_energy:word;
	  old_pitch:byte;
	  old_k:array[0..10-1] of integer;
	  target_energy:word;
	  target_pitch:byte;
	  target_k:array[0..10-1] of integer;

	  new_energy:word;
	  new_pitch:byte;
	  new_k:array[0..10-1] of integer;

	  // these are all used to contain the current state of the sound generation */
	  current_energy:cardinal;
	  current_pitch:cardinal;
	  current_k:array[0..10-1] of integer;
	  x:array[0..10-1] of integer;
    amplificador:byte;
  end;
  pvlm5030=^vlm5030_info;
var
  chip_sn:array[0..3] of pvlm5030;

// get BSY pin level */
function VLM5030_BSY(numero:byte):integer;pascal;
// latch contoll data */
procedure VLM5030_data_w(numero,data:byte);
// set RST pin level : reset / set table address A8-A15 */
procedure VLM5030_RST(numero:byte;pin:integer);
// set VCU pin level : ?? unknown */
procedure VLM5030_VCU(numero:byte;pin:integer);
// set ST pin level  : set table address A0-A7 / start speech */
procedure VLM5030_ST(numero:byte;pin:integer);
procedure vlm5030_start(numero,amplificador:byte;clock:integer;prom:pbyte;rom_size:integer);
procedure vlm5030_update(numero,canal:byte);
procedure vlm5030_close(numero:byte);
procedure VLM5030_reset(numero:byte);

implementation

function get_bits(numero:byte;sbit,bits:integer):integer;
var
  data,offset:integer;
  chip:pvlm5030;
  p:pbyte;
begin
  chip:=chip_sn[numero];
	offset:= chip.address + (sbit shr 3);

  p:=chip.rom;
  inc(p,(offset and chip.address_mask));
	data:=p^;
  p:=chip.rom;
  inc(p,((offset+1) and chip.address_mask));
  data:=data+(p^*256);

	data:=data shr (sbit and 7);
	data:=data and ($ff shr (8-bits));

	get_bits:=data;
end;

// get next frame */
function parse_frame(numero:byte):integer;
var
  cmd:byte;
	i,nums:integer;
  chip:pvlm5030;
  p:pbyte;
begin
  chip:=chip_sn[numero];
	// remember previous frame */
	chip.old_energy:= chip.new_energy;
	chip.old_pitch:= chip.new_pitch;
	for i:=0 to 9 do chip.old_k[i]:= chip.new_k[i];
  // command byte check */
  p:=chip.rom;
  inc(p,(chip.address and chip.address_mask));
  cmd:=p^;
  if (cmd and $01)<>0 then begin
	    // extend frame */
		  chip.new_energy:=0;
      chip.new_pitch:=0;
		  for i:=0 to 9 do chip.new_k[i]:=0;
		  inc(chip.address);
      if (cmd and $02 )<>0 then begin
			  // end of speech */
			  // logerror("VLM5030 %04X end \n",chip->address ); */
			  parse_frame:=0;
        exit;
		  end else begin
			  // silent frame */
			  nums:= ( (cmd shr 2)+1 )*2;
			  // logerror("VLM5030 %04X silent %d frame\n",chip->address,nums ); */
			  parse_frame:=nums * FR_SIZE;
        exit;
		  end;
	end;
	// pitch */
	chip.new_pitch:=(pitchtable[get_bits(numero, 1,5)] + chip.pitch_offset ) and $ff;
	// energy */
	chip.new_energy:= energytable[get_bits(numero, 6,5)];

	// 10 K's */
	chip.new_k[9]:= K5_table[get_bits(numero,11,3)];
	chip.new_k[8]:= K5_table[get_bits(numero,14,3)];
	chip.new_k[7]:= K5_table[get_bits(numero,17,3)];
	chip.new_k[6]:= K5_table[get_bits(numero,20,3)];
	chip.new_k[5]:= K5_table[get_bits(numero,23,3)];
	chip.new_k[4]:= K5_table[get_bits(numero,26,3)];
	chip.new_k[3]:= K3_table[get_bits(numero,29,4)];
	chip.new_k[2]:= K3_table[get_bits(numero,33,4)];
	chip.new_k[1]:= K2_table[get_bits(numero,37,5)];
	chip.new_k[0]:= K1_table[get_bits(numero,42,6)];

	inc(chip.address,6);
	//logerror("VLM5030 %04X voice \n",chip->address );
	parse_frame:=FR_SIZE;
end;

// decode and buffering data */
{Para complensar el sampleado bajo del vlm 5030, meto el valor
varias veces. Por ejemplo 3580000/440=8136,6 sampleados/seg para
compensar hasta los 44100, 44100/8136,6=5,42 veces mas, pongo el
valor obtenido 6 veces en el sample}
procedure vlm5030_update(numero,canal:byte);
var
	interp_effect,i,current_val:integer;
	u:array[0..11-1] of integer;
  chip:pvlm5030;
  val:smallint;
label
  phase_stop;
begin
  chip:=chip_sn[numero];
	// running */
	if ((chip.phase=PH_RUN) or (chip.phase=PH_STOP)) then begin
		// playing speech */
		//while (length > 0) do begin
			// check new interpolator or  new frame */
			if (chip.sample_count=0) then begin
				if (chip.phase=PH_STOP) then begin
					chip.phase:=PH_END;
					chip.sample_count:= 1;
					goto phase_stop; // continue to end phase */
				end;
				chip.sample_count:= chip.frame_size;
				// interpolator changes */
				if (chip.interp_count=0) then begin
					// change to new frame */
					chip.interp_count:= parse_frame(numero); // with change phase */
					if (chip.interp_count=0) then begin
					 	// end mark found */
						chip.interp_count:= FR_SIZE;
						chip.sample_count:= chip.frame_size; // end -> stop time */
						chip.phase:= PH_STOP;
					end;
					// Set old target as new start of frame */
					chip.current_energy:= chip.old_energy;
					chip.current_pitch:= chip.old_pitch;
					for i:=0 to 9 do chip.current_k[i]:=chip.old_k[i];
					// is this a zero energy frame? */
					if (chip.current_energy=0) then begin
						//mame_printf_debug("processing frame: zero energy\n");*/
						chip.target_energy:= 0;
						chip.target_pitch:= chip.current_pitch;
						for i:=0 to 9 do chip.target_k[i]:=chip.current_k[i];
					end else begin
						chip.target_energy:= chip.new_energy;
						chip.target_pitch:= chip.new_pitch;
						for i:=0 to 9 do chip.target_k[i]:=chip.new_k[i];
					end;
				end;
				// next interpolator */
				// Update values based on step values 25% , 50% , 75% , 100% */
				dec(chip.interp_count,chip.interp_step);
				// 3,2,1,0 -> 1,2,3,4 */
				interp_effect:= FR_SIZE - (chip.interp_count mod FR_SIZE);
				chip.current_energy:= chip.old_energy + (chip.target_energy - chip.old_energy) * interp_effect div FR_SIZE;
				if (chip.old_pitch>1) then chip.current_pitch:= chip.old_pitch + (chip.target_pitch - chip.old_pitch) * interp_effect div FR_SIZE;
				for i:=0 to 9 do chip.current_k[i]:=chip.old_k[i] + (chip.target_k[i] - chip.old_k[i]) * interp_effect div FR_SIZE;
			end;
			// calcrate digital filter */
			if (chip.old_energy=0) then begin
				// generate silent samples here */
				current_val:=$00;
			end else if (chip.old_pitch <= 1) then begin
				  // generate unvoiced samples here */
          if (random(256) and 1)<>0 then current_val:=chip.current_energy
            else current_val:=-chip.current_energy;
			  end else begin
				  // generate voiced samples here */
				  if (chip.pitch_count=0) then current_val:=chip.current_energy
            else current_val:=0;
			  end;

			// Lattice filter here */
			u[10]:=current_val;
			for i:=9 downto 0 do u[i]:=u[i+1]-((chip.current_k[i] * chip.x[i]) div 32768);
			for i:=9 downto 1 do chip.x[i]:=chip.x[i-1] + ((chip.current_k[i-1] * u[i-1]) div 32768);
			chip.x[0] :=u[0];

			// clipping, buffering */
			if (u[0] > 511) then val:= 511 shl 6
			  else if (u[0] < -511) then val:=-511 shl 6
			else val:=u[0] shl 6;
      for i:=0 to 5 do tsample[canal][posicion_sonido+i]:=val*chip.amplificador;
			// sample count */
			dec(chip.sample_count);
			// pitch */
			inc(chip.pitch_count);
			if (chip.pitch_count >= chip.current_pitch ) then chip.pitch_count:=0;
		//end; //del while
      exit;
	end;
	// stop phase */
phase_stop:
	case (chip.phase) of
	  PH_SETUP:if (chip.sample_count<=1) then begin
			        chip.sample_count:=0;
			        // logerror("VLM5030 BSY=H\n" ); */
			        // pin_BSY = 1; */
			        chip.phase:=PH_WAIT;
		         end else begin
			        dec(chip.sample_count);
		         end;
	  PH_END:if (chip.sample_count<=1) then begin
			      chip.sample_count:= 0;
			      // logerror("VLM5030 BSY=L\n" ); */
			      chip.pin_BSY:= 0;
			      chip.phase:= PH_IDLE;
		      end	else begin
			      dec(chip.sample_count);
		      end;
	end;
	// silent buffering
  for i:=0 to 5 do tsample[canal][posicion_sonido+i]:=0;
end;

// get BSY pin level */
function VLM5030_BSY(numero:byte):integer;pascal;
var
  chip:pvlm5030;
begin
  chip:=chip_sn[numero];
	VLM5030_BSY:=chip.pin_BSY;
end;

// latch contoll data */
procedure VLM5030_data_w(numero,data:byte);
var
  chip:pvlm5030;
begin
  chip:=chip_sn[numero];
	chip.latch_data:=data;
end;

// setup parameteroption when RST=H */
procedure VLM5030_setup_parameter(numero,param:byte);
var
  chip:pvlm5030;
begin
  chip:=chip_sn[numero];
	// latch parameter value */
	chip.parameter:=param;

	// bit 0,1 : 4800bps / 9600bps , interporator step */
	if (param and 2)<>0 then // bit 1 = 1 , 9600bps */
		chip.interp_step:= 4 // 9600bps : no interporator */
	else if(param and 1)<>0 then // bit1 = 0 & bit0 = 1 , 4800bps */
		chip.interp_step:= 2 // 4800bps : 2 interporator */
	else	// bit1 = bit0 = 0 : 2400bps */
		chip.interp_step:= 1; // 2400bps : 4 interporator */

	// bit 3,4,5 : speed (frame size) */
	chip.frame_size:=VLM5030_speed_table[(param shr 3) and 7];

	// bit 6,7 : low / high pitch */
	if (param and $80)<>0 then	// bit7=1 , high pitch */
		chip.pitch_offset:= -8
	else if (param and $40)<>0 then	// bit6=1 , low pitch */
		chip.pitch_offset:= 8
	else
		chip.pitch_offset:= 0;
end;

procedure VLM5030_reset(numero:byte);
var
  chip:pvlm5030;
begin
  chip:=chip_sn[numero];

	chip.phase:= PH_RESET;
	chip.address:= 0;
	chip.vcu_addr_h:= 0;
	chip.pin_BSY:= 0;

	chip.old_energy:=0;
  chip.old_pitch:=0;
	chip.new_energy:=0;
  chip.new_pitch:=0;
	chip.current_energy:=0;
  chip.current_pitch:= 0;
	chip.target_energy:=0;
  chip.target_pitch:= 0;

	zeromemory(@chip.old_k[0],10*4);
	zeromemory(@chip.new_k[0],10*4);
	zeromemory(@chip.current_k,10*4);
	zeromemory(@chip.target_k,10*4);
	chip.interp_count:=0;
  chip.sample_count:=0;
  chip.pitch_count:= 0;
	zeromemory(@chip.x[0],10*4);
	// reset parameters */
	VLM5030_setup_parameter(numero,$00);
end;


// set RST pin level : reset / set table address A8-A15 */
procedure VLM5030_RST(numero:byte;pin:integer);
var
  chip:pvlm5030;
begin
  chip:=chip_sn[numero];
	if chip.pin_RST<>0 then begin
		if (pin=0) then begin
			// H -> L : latch parameters */
			chip.pin_RST:= 0;
			VLM5030_setup_parameter(numero,chip.latch_data);
		end;
	end else begin
		if (pin<>0) then begin
			// L -> H : reset chip */
			chip.pin_RST:= 1;
			if chip.pin_BSY<>0 then begin
				VLM5030_reset(numero);
			end;
		end;
	end;
end;

// set VCU pin level : ?? unknown */
procedure VLM5030_VCU(numero:byte;pin:integer);
var
  chip:pvlm5030;
begin
  chip:=chip_sn[numero];
	// direct mode / indirect mode */
	chip.pin_VCU:= pin;
end;

// set ST pin level  : set table address A0-A7 / start speech */
procedure VLM5030_ST(numero:byte;pin:integer);
var
  chip:pvlm5030;
  table:integer;
  p:pbyte;
begin
  chip:=chip_sn[numero];
	if (chip.pin_ST<>pin) then begin
		// pin level is change */
		if (pin=0) then begin
			// H -> L */
			chip.pin_ST:= 0;
			if( chip.pin_VCU<>0 ) then begin
				// direct access mode & address High */
				chip.vcu_addr_h:=(chip.latch_data shl 8) + $01;
			end	else begin
				// start speech */
				// check access mode */
				if (chip.vcu_addr_h<>0) then begin
					// direct access mode */
					chip.address:= (chip.vcu_addr_h and $ff00) + chip.latch_data;
					chip.vcu_addr_h:= 0;
				end	else begin
					// indirect accedd mode */
					table:= (chip.latch_data and $fe) + ((chip.latch_data and 1) shl 8);
          p:=chip.rom;
          inc(p,table and chip.address_mask);
					chip.address:=(p^ shl 8);
          p:=chip.rom;
          inc(p,(table+1) and chip.address_mask);
          chip.address:=chip.address or p^;
				end;
				// logerror("VLM5030 %02X start adr=%04X\n",table/2,chip->address ); */
				// reset process status */
				chip.sample_count:= chip.frame_size;
				chip.interp_count:= FR_SIZE;
				// clear filter */
				// start after 3 sampling cycle */
				chip.phase:= PH_RUN;
			end;
		end	else begin
			// L -> H */
			chip.pin_ST:= 1;
			// setup speech , BSY on after 30ms? */
			chip.phase:= PH_SETUP;
			chip.sample_count:= 1; // wait time for busy on */
			chip.pin_BSY:= 1;
		end;
	end;
end;

// start VLM5030 with sound rom              */
procedure vlm5030_start(numero,amplificador:byte;clock:integer;prom:pbyte;rom_size:integer);
var
  chip:pvlm5030;
begin
  if chip_sn[numero]=nil then begin
    getmem(chip_sn[numero],sizeof(vlm5030_info));
    zeromemory(chip_sn[numero],sizeof(vlm5030_info));
  end;
  chip:=chip_sn[numero];

	//emulation_rate:= clock / 440;

	// reset input pins */
	chip.pin_RST:=0;
  chip.pin_ST:=0;
  chip.pin_VCU:= 0;
	chip.latch_data:= 0;

	VLM5030_reset(numero);
	chip.phase:= PH_IDLE;

	chip.rom:=prom;
  chip.amplificador:=amplificador;
	// memory size */
	chip.address_mask:=rom_size-1;
end;

procedure vlm5030_close(numero:byte);
begin
  if chip_sn[numero]<>nil then begin
    freemem(chip_sn[numero]);
    chip_sn[numero]:=nil;
  end;
end;

end.
