unit ym_2203;

interface
uses fmopn,ay_8910,windows,timers;

const
  MAX_YM2203=3;
type
  YM2203_f=record
     OPN:pfm_opn;
     CH:array[0..3] of pfm_chan;
     REGS:array[0..255] of byte;
     AY_amp:smallint;
     timer_enabled:boolean;
  end;
  pYM2203_f=^YM2203_f;
  IRQ_Handler=procedure (numchip:integer;irqstate:integer);pascal;

var
  FM2203:array[0..MAX_YM2203] of pYM2203_f;

procedure ym2203_init_timers(sound_clock:integer;num_snd_cpu:byte;num:byte;irq_func:IRQ_Handler);
procedure ym2203_init(num,clock:integer;ay_amp:smallint;porta_read,portb_read:read_port;porta_write,portb_write:write_port);
procedure ym2203_close(num:byte);
function ym2203_read_status(num:integer):integer;
function ym2203_read_reg(num:integer):integer;
procedure ym2203_control(num,data:integer);
procedure ym2203_write_reg(num,data:integer);
function YM2203UpdateOne(chip:integer):integer;
procedure YM2203_Reset(num:integer);
procedure timer1;pascal;

implementation
uses principal;

procedure reset_channels(ST:pfm_state;chip:integer;num:integer);
var
  c,s:integer;
  F2203:pYM2203_f;
begin
  F2203:=FM2203[chip];
	ST.mode  := 0;	// normal mode */
	ST.TA    := 0;
	ST.TAC   := 0;
	ST.TB    := 0;
	ST.TBC   := 0;
	for c:=0 to num-1 do begin
		F2203.CH[c].fc:= 0;
		for s:=0 to 3 do begin
			F2203.CH[c].SLOT[s].ssg:= 0;
			F2203.CH[c].SLOT[s].ssgn:= 0;
			F2203.CH[c].SLOT[s].state:= EG_OFF;
			F2203.CH[c].SLOT[s].volume:= MAX_ATT_INDEX;
			F2203.CH[c].SLOT[s].vol_out:= MAX_ATT_INDEX;
		end;
	end;
end;

procedure YM2203_Reset(num:integer);
var
  i:integer;
  OPN:pfm_opn;
  F2203:pYM2203_f;
begin
    F2203:=FM2203[num];
		OPN:=F2203.OPN;
		// Reset Prescaler
	  OPNPrescaler_w(OPN, 0 , 1 );
	  // reset SSG section */
	  AY8910_reset(num);
	  // status clear */
	  FM_IRQMASK_SET(OPN.ST,$03);
	  OPNWriteMode(OPN,$27,$30); // mode 0 , timer reset */
	  OPN.eg_timer := 0;
	  OPN.eg_cnt   := 0;
	  FM_STATUS_RESET(OPN.ST,$ff);
	  reset_channels(OPN.ST,num,3);
		// reset OPerator paramater */
		for i:=$b2 downto $30 do OPNWriteReg(OPN,i,0);
		for i:=$26 downto $20 do OPNWriteReg(OPN,i,0);
end;

procedure ym2203_close(num:byte);
var
  i:byte;
begin
//El PSG
AY8910_close(num);
//cierro los chans
for i:=0 to 3 do begin
  fmchan_close(FM2203[num].CH[i]);
  FM2203[num].OPN.P_CH[i]:=nil;
end;
//Cierro el OPN
opn_close(FM2203[num].OPN);
if FM2203[num]<>nil then begin
  freemem(FM2203[num],sizeof(ym2203_f));
  FM2203[num]:=nil;
end;
end;

procedure ym2203_init_timers(sound_clock:integer;num_snd_cpu:byte;num:byte;irq_func:IRQ_Handler);
begin
  timer[5].cpu:=num_snd_cpu;
  //El YM2203 tiene 2 temporizadores programables, hacen un tic
  //cada 200useg
  //El segundo factor es la sincronizacion proporcional entre la velocidad de
  //la cpu de audio y el YM2203.
  timer[5].tiempo:=(sound_clock*0.00002)*(FM2203[num].OPN.ST.clock/sound_clock);
  timer[5].ejecutar:=timer1;
  timer[5].enabled:=true;
  FM2203[num].timer_enabled:=true;
  FM2203[num].OPN.ST.IRQ_Handler:=irq_func;
end;

procedure ym2203_init(num,clock:integer;ay_amp:smallint;porta_read,portb_read:read_port;porta_write,portb_write:write_port);
var
  i:byte;
begin
  //YM2203NumChips:=num;
  FMInitTable;
  if FM2203[num]=nil then begin
      getmem(FM2203[num],sizeof(ym2203_f));
      zeromemory(FM2203[num],SizeOf(ym2203_f));
  end;
  //El PSG
  AY8910_init(clock,num,porta_read,portb_read,porta_write,portb_write);
  //Inicializo el OPN
  FM2203[num].OPN:=opn_init;
  //inico los otros chans
  for i:=0 to 3 do begin
    FM2203[num].CH[i]:=fmchan_init;
    FM2203[num].OPN.P_CH[i]:=FM2203[num].CH[i];
  end;
  //Inicializo el state
  FM2203[num].OPN.type_:=TYPE_YM2203;
  FM2203[num].OPN.ST.clock:=clock;
  FM2203[num].OPN.ST.rate:=44100;
  FM2203[num].AY_amp:=ay_amp;
  FM2203[num].timer_enabled:=false;
  YM2203_Reset(num);
end;

function YM2203UpdateOne(chip:integer):integer;
var
  F2203:pYM2203_f;
  OPN:pfm_opn;
  lt:integer;
  cch:array[0..2] of pfm_chan;
begin
		F2203:=FM2203[chip];
		OPN:=F2203.OPN;

		cch[0]:=F2203.CH[0];
		cch[1]:=F2203.CH[1];
		cch[2]:=F2203.CH[2];

    // refresh PG and EG */
	  refresh_fc_eg_chan(OPN,cch[0]);
	  refresh_fc_eg_chan(OPN,cch[1]);

		if ((F2203.OPN.ST.mode and $c0)<>0) then begin
		  // 3SLOT MODE */
		  if (cch[2].SLOT[SLOT1].Incr=-1) then begin
			  refresh_fc_eg_slot(OPN,cch[2].SLOT[SLOT1],OPN.SL3.fc[1],OPN.SL3.kcode[1] );
			  refresh_fc_eg_slot(OPN,cch[2].SLOT[SLOT2],OPN.SL3.fc[2],OPN.SL3.kcode[2] );
			  refresh_fc_eg_slot(OPN,cch[2].SLOT[SLOT3],OPN.SL3.fc[0],OPN.SL3.kcode[0] );
			  refresh_fc_eg_slot(OPN,cch[2].SLOT[SLOT4],cch[2].fc,cch[2].kcode );
		  end;
    end else refresh_fc_eg_chan(OPN,cch[2]);

	  // YM2203 doesn't have LFO so we must keep these globals at 0 level */
	  LFO_AM:= 0;
	  LFO_PM:= 0;

	  // buffering */
	  //for (i=0; i < length ; i++) do begin
		  // clear outputs */
		  out_fm[0]:= 0;
		  out_fm[1]:= 0;
		  out_fm[2]:= 0;

		  // advance envelope generator */
		  OPN.eg_timer:=OPN.eg_timer+OPN.eg_timer_add;
		  while (OPN.eg_timer>=OPN.eg_timer_overflow) do begin
			  OPN.eg_timer:=OPN.eg_timer-OPN.eg_timer_overflow;
			  OPN.eg_cnt:=OPN.eg_cnt+1;
			  advance_eg_channel(OPN,cch[0]);
			  advance_eg_channel(OPN,cch[1]);
			  advance_eg_channel(OPN,cch[2]);
		  end;

		  // calculate FM */
		  chan_calc(OPN,cch[0],0);
		  chan_calc(OPN,cch[1],0);
		  chan_calc(OPN,cch[2],0);

		  // buffering */
			lt:=out_fm[0]+out_fm[1]+out_fm[2]+round(renderbyte(chip)^*F2203.AY_amp);
      if lt>$7fff then lt:=$7fff;
      if lt<-$7fff then lt:=-$7fff;
			YM2203UpdateOne:=lt;

		// timer A control */
		//INTERNAL_TIMER_A(F2203.OPN.ST,F2203.CH[2]);
	//end;
	//INTERNAL_TIMER_B(F2203.OPN.ST,1)
end;

function YM2203Write(n,a,v:integer):integer;
var
  OPN:pfm_opn;
  addr:integer;
  F2203:pYM2203_f;
begin
    F2203:=FM2203[n];
		OPN:=F2203.OPN;
		if ((a and 1)= 0 ) then begin
			// address port */
      v:=v and $ff;
			OPN.ST.address:=v;
			// Write register to SSG emurator */
			if (v<16) then AYControl(v,n);
			// prescaler select : 2d,2e,2f  */
		  if ((v>=$2d) and (v<=$2f)) then OPNPrescaler_w(OPN,v,1);
		end else begin
			// data port */
			addr:=OPN.ST.address;
      F2203.REGS[addr]:=v;
			case (addr and $f0) of
			  $00:begin	// 0x00-0x0f : SSG section */
              // Write data to SSG emurator */
              AYWrite(v,n);
            end;
			  $20:begin	// 0x20-0x2f : Mode section */
				      //YM2203UpdateReq(n);
				      // write register */
				      OPNWriteMode(OPN,addr,v);
				    end;
			  else begin	// 0x30-0xff : OPN section */
				      //YM2203UpdateReq(n);
				      // write register */
				      OPNWriteReg(OPN,addr,v);
              end;
		  end;
    end;
		YM2203Write:=OPN.ST.irq;
end;

function YM2203Read(n,a:integer):integer;
var
  F2203:pym2203_f;
  adr:integer;
  ret:byte;
begin
		F2203:=FM2203[n];
		adr:= F2203.OPN.ST.address;
		ret:= 0;

		if ((a and 1)=0 ) then begin
			// status port */
			ret:=F2203.OPN.ST.status;
    end else begin
			// data port (ONLY SSG) */
			if (adr<16) then begin
        ret:=AYRead(n);
      end;
		end;
		YM2203Read:=ret;
end;

function ym2203_read_status(num:integer):integer;
begin
  ym2203_read_status:=YM2203Read(num,0);
end;

function ym2203_read_reg(num:integer):integer;
begin
  ym2203_read_reg:=YM2203Read(num,1);
end;

procedure ym2203_control(num,data:integer);
begin
  YM2203Write(num,0,data);
end;

procedure ym2203_write_reg(num,data:integer);
begin
  YM2203Write(num,1,data);
end;

procedure timer1;pascal;
var
  f:byte;
begin
for f:=0 to MAX_YM2203 do
  if ((FM2203[f]<>nil) and FM2203[f].timer_enabled) then begin
    INTERNAL_TIMER_A(FM2203[f].OPN.ST,FM2203[f].CH[2]);
    INTERNAL_TIMER_B(FM2203[f].OPN.ST,1)
  end;
end;

end.
