unit ym_3812;

interface
uses fmopl,windows,timers;

const
  MAX_YM3812=3;

type
  YM3812_f=record
     OPL:pfm_opl;
     timer_enabled:boolean;
  end;
  pYM3812_f=^YM3812_f;
  IRQ_Handler=procedure (numchip:integer;irqstate:integer);pascal;
var
  FM3812:array[0..MAX_YM3812] of pYM3812_f;

function YM3812UpdateOne(num:byte):integer;
procedure YM3812_Init(num:byte;clock:dword);
procedure YM3812_ResetChip(num:byte);
procedure YM3812_Close(num:byte);
procedure YM3812_control_port(num,data:byte);
procedure YM3812_write_port(num,data:byte);
function YM3812_status_port(num:byte):integer;
function YM3812_read_port(num:byte):integer;
procedure ym3812_init_timers(sound_clock:integer;num_snd_cpu:byte;num:byte;irq_func:IRQ_Handler);pascal;
procedure timer1;pascal;
procedure timer2;pascal;

implementation

procedure ym3812_init_timers(sound_clock:integer;num_snd_cpu:byte;num:byte;irq_func:IRQ_Handler);pascal;
begin
  //El YM3812 tiene 2 temporizadores programables, uno hace un tic
  //cada 80useg y el otro cada 320useg
  //El segundo factor es la sincronizacion proporcional entre la velocidad de
  //la cpu de audio y el YM3812.
  timer[6].cpu:=num_snd_cpu;
  timer[6].tiempo:=(sound_clock*0.0000175)*(FM3812[num].OPL.clock/sound_clock);
  timer[6].ejecutar:=timer1;
  timer[6].enabled:=true;
  timer[7].cpu:=num_snd_cpu;
  timer[7].tiempo:=(sound_clock*0.00007)*(FM3812[num].OPL.clock/sound_clock);
  timer[7].ejecutar:=timer2;
  timer[7].enabled:=true;
  FM3812[num].timer_enabled:=true;
  FM3812[num].OPL.IRQ_Handler:=irq_func;
end;

procedure YM3812_Init(num:byte;clock:dword);
var
  rate:integer;
begin
  rate:=round(clock/72);
	// emulator create */
  if FM3812[num]=nil then begin
    getmem(FM3812[num],sizeof(ym3812_f));
    zeromemory(FM3812[num],SizeOf(ym3812_f));
  end;
  FM3812[num].OPL:=OPLCreate(clock,rate);
  YM3812_ResetChip(num);
  FM3812[num].timer_enabled:=false;
end;

procedure YM3812_close(num:byte);
var
  OPL:pfm_opl;
begin
	OPL:=FM3812[num].OPL;
	OPLClose(OPL);
  freemem(FM3812[num]);
  FM3812[num]:=nil;
end;

procedure YM3812_ResetChip(num:byte);
var
  OPL:pfm_opl;
begin
	OPL:=FM3812[num].OPL;
	OPLResetChip(OPL);
  cur_chip:=$ff;
end;

{procedure YM3812Shutdown(num:byte);
var
  OPL:pfm_opl;
begin
	OPL:=FM3812[num].OPL;
	// emulator shutdown */
	OPLDestroy(OPL);
end;}

//
// Generate samples for one of the YM3812's
//

function YM3812UpdateOne(num:byte):integer;
var
	OPL:pFM_OPL;
	rhythm:boolean;
  lt:integer;
begin
  OPL:=FM3812[num].OPL;
  rhythm:=(OPL.rhythm and $20)<>0;
	if (num<>cur_chip) then begin
		cur_chip:=num;
		// rhythm slots */
		SLOT7_1:=OPL.P_CH[7].SLOT[SLOT1];
		SLOT7_2:=OPL.P_CH[7].SLOT[SLOT2];
		SLOT8_1:=OPL.P_CH[8].SLOT[SLOT1];
		SLOT8_2:=OPL.P_CH[8].SLOT[SLOT2];
	end;
  output:= 0;
	advance_lfo(OPL);
  // FM part */
  OPL_CALC_CH(OPL.P_CH[0]);
  OPL_CALC_CH(OPL.P_CH[1]);
  OPL_CALC_CH(OPL.P_CH[2]);
  OPL_CALC_CH(OPL.P_CH[3]);
  OPL_CALC_CH(OPL.P_CH[4]);
  OPL_CALC_CH(OPL.P_CH[5]);
  if not(rhythm) then begin
			OPL_CALC_CH(OPL.P_CH[6]);
			OPL_CALC_CH(OPL.P_CH[7]);
			OPL_CALC_CH(OPL.P_CH[8]);
  end else begin		// Rhythm part */
			OPL_CALC_RH(OPL.P_CH[6], (OPL.noise_rng shr 0) and 1 );
  end;

  lt:=output shl 1;

  //lt:=lt shr FINAL_SH;

  // limit check */
  if lt>$7fff then lt:=$7fff;
  if lt<-$7fff then lt:=-$7fff;

  YM3812UpdateOne:=lt;

  //INTERNAL_TIMER_A(FM3812[num].OPL);
  //INTERNAL_TIMER_B(FM3812[num].OPL);

  // store to sound buffer */
  advance(OPL);
end;

procedure YM3812_control_port(num,data:byte);
begin
  FM3812[num].OPL.address:=data;
end;

procedure YM3812_write_port(num,data:byte);
begin
		OPLWriteReg(FM3812[num].OPL,FM3812[num].OPL.address,data);
end;

function YM3812_status_port(num:byte):integer;
begin
	YM3812_status_port:=(FM3812[num].OPL.status and (FM3812[num].OPL.statusmask or $80)) or $06 ;
end;

function YM3812_read_port(num:byte):integer;
begin
	YM3812_read_port:=$ff;
end;

procedure timer1;pascal;
var
  f:byte;
begin
for f:=0 to MAX_YM3812 do
  if ((FM3812[f]<>nil) and FM3812[f].timer_enabled) then
    INTERNAL_TIMER_A(FM3812[f].OPL);
end;

procedure timer2;pascal;
var
  f:byte;
begin
for f:=0 to MAX_YM3812 do
  if ((FM3812[f]<>nil) and FM3812[f].timer_enabled) then
  INTERNAL_TIMER_B(FM3812[f].OPL);
end;


end.
